@@ -3,12 +3,16 @@ module Main where
 import Data.Monoid ((<>))
 import System.FilePath ((</>))
 
+import qualified GHC.IO.Encoding
 import qualified System.Directory
+import qualified System.IO
 import qualified Test.Mockery.Directory
 import qualified Test.DocTest
 
 main :: IO ()
 main = do
+   
+    GHC.IO.Encoding.setLocaleEncoding System.IO.utf8 
     pwd    <- System.Directory.getCurrentDirectory
     prefix <- System.Directory.makeAbsolute pwd
 
@@ -21,8 +25,10 @@ main = do
         writeFile "file1" "./file2"
 
         Test.DocTest.doctest
-            [ "--fast"
+            [ "-DWITH_HTTP"
+            , "--fast"
             , "-i" <> (prefix </> "src")
+            , "-i" <> (prefix </> "ghc-src")
             , prefix </> "src/Dhall.hs"
             , prefix </> "src/Dhall/Import.hs"
             , prefix </> "src/Dhall/Tutorial.hs"
