@@ -12,21 +12,21 @@
 module Dhall.Diff (
     -- * Diff
       Diff (..)
-    , diffExpression
     , diffNormalized
-    , Dhall.Diff.diff
+    , diff
     ) where
 
 import Data.Foldable (fold, toList)
-import Data.Function (on)
 import Data.List.NonEmpty (NonEmpty(..))
 import Data.Monoid (Any(..))
-import Data.Semigroup
+import Data.Semigroup hiding (diff)
 import Data.Sequence (Seq)
 import Data.String (IsString(..))
 import Data.Text (Text)
 import Data.Text.Prettyprint.Doc (Doc, Pretty)
+import Data.Void (Void)
 import Dhall.Core (Binding(..), Chunks (..), Const(..), Expr(..), Var(..))
+import Dhall.Binary (ToTerm)
 import Dhall.Map (Map)
 import Dhall.Set (Set)
 import Dhall.Pretty.Internal (Ann)
@@ -153,18 +153,12 @@ rparen :: Diff
 rparen = token Internal.rparen
 
 -- | Render the difference between the normal form of two expressions
-diffNormalized :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Doc Ann
+diffNormalized :: (Eq a, Pretty a, ToTerm a) => Expr s a -> Expr s a -> Diff
 diffNormalized l0 r0 = Dhall.Diff.diff l1 r1
   where
     l1 = Dhall.Core.alphaNormalize (Dhall.Core.normalize l0)
     r1 = Dhall.Core.alphaNormalize (Dhall.Core.normalize r0)
 
--- | Render the difference between two expressions
-diff :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Doc Ann
-diff l0 r0 = doc
-  where
-    Diff {..} = diffExpression l0 r0
-
 diffPrimitive :: Eq a => (a -> Diff) -> a -> a -> Diff
 diffPrimitive f l r
     | l == r    = ignore
@@ -206,12 +200,16 @@ diffBool = diffPrimitive bool
 diffInteger :: Integer -> Integer -> Diff
 diffInteger = diffPrimitive (token . Internal.prettyNumber)
 
+diffInt :: Int -> Int -> Diff
+diffInt = diffPrimitive (token . Internal.prettyInt)
+
 diffVar :: Var -> Var -> Diff
-diffVar (V xL nL) (V xR nR) = format mempty label <> "@" <> natural
+diffVar (V xL nL) (V xR nR) =
+    format mempty label <> if same natural then mempty else "@" <> natural
   where
     label = diffLabel xL xR
 
-    natural = diffInteger nL nR
+    natural = diffInt nL nR
 
 diffPretty :: (Eq a, Pretty a) => a -> a -> Diff
 diffPretty = diffPrimitive (token . Pretty.pretty)
@@ -252,14 +250,22 @@ enclosed' l m docs =
 diffKeyVals
     :: (Eq a, Pretty a)
     => Diff
-    -> Map Text (Expr s a)
-    -> Map Text (Expr s a)
+    -> Map Text (Expr Void a)
+    -> Map Text (Expr Void a)
     -> [Diff]
-diffKeyVals assign kvsL kvsR =
+diffKeyVals assign = diffKeysWith assign diff
+
+diffKeysWith
+    :: Diff
+    -> (a -> a -> Diff)
+    -> Map Text a
+    -> Map Text a
+    -> [Diff]
+diffKeysWith assign diffVals kvsL kvsR =
     diffFieldNames <> diffFieldValues <> (if anyEqual then [ ignore ] else [])
   where
-    ksL = Data.Set.fromList (Dhall.Map.keys kvsL)
-    ksR = Data.Set.fromList (Dhall.Map.keys kvsR)
+    ksL = Dhall.Map.keysSet kvsL
+    ksR = Dhall.Map.keysSet kvsR
 
     extraL = Data.Set.difference ksL ksR
     extraR = Data.Set.difference ksR ksL
@@ -274,7 +280,7 @@ diffKeyVals assign kvsL kvsR =
             <>  ignore
             ]
 
-    shared = Dhall.Map.intersectionWith diffExpression kvsL kvsR
+    shared = Dhall.Map.intersectionWith diffVals kvsL kvsR
 
     diffFieldValues =
         filter (not . same) (Dhall.Map.foldMapWithKey adapt shared)
@@ -300,15 +306,8 @@ bracketed :: [Diff] -> Diff
 bracketed = enclosed (lbracket <> " ") (comma <> " ") rbracket
 
 diffText :: Text -> Text -> Diff
-diffText l r
-  | null parts         = "\"\""
-  | allDifferent parts = difference textSkeleton textSkeleton
-  | allSame parts      = textSkeleton
-  | otherwise          = "\"" <> foldMap prettyPart parts <> "\""
+diffText l r = "\"" <> foldMap prettyPart parts <> "\""
   where
-    allDifferent = not . any isBoth
-    allSame      = all isBoth
-
     -- TODO: check for color support from the TTY
     colorDiff colorCode chars =
             "\ESC["
@@ -338,7 +337,7 @@ diffText l r
 
 diffChunks
     :: (Eq a, Pretty a)
-    => Chunks s a -> Chunks s a -> Diff
+    => Chunks Void a -> Chunks Void a -> Diff
 diffChunks cL cR
   | null chunks             = "\"\""
   | [c] <- chunks           = c
@@ -354,21 +353,17 @@ diffChunks cL cR
     chunkDiff a b =
       case (a, b) of
         (Left  x, Left y ) -> diffText x y
-        (Right x, Right y) -> diffExpression x y
+        (Right x, Right y) -> diff x y
         _                  -> diffTextSkeleton
 
 diffList
     :: (Eq a, Pretty a)
-    => Seq (Expr s a) -> Seq (Expr s a) -> Diff
-diffList l r
-  | allDifferent parts = difference listSkeleton listSkeleton
-  | otherwise          = bracketed (foldMap diffPart parts)
+    => Seq (Expr Void a) -> Seq (Expr Void a) -> Diff
+diffList l r = bracketed (foldMap diffPart parts)
   where
-    allDifferent = any (not . isBoth)
-
     -- Sections of the list that are only in left, only in right, or in both
     parts =
-        Algo.Diff.getGroupedDiffBy ((same .) . diffExpression) (toList l) (toList r)
+        Algo.Diff.getGroupedDiffBy ((same .) . diff) (toList l) (toList r)
 
     -- Render each element of a list using an extra rendering function f
     prettyElems f = map (f . token . Internal.prettyExpr)
@@ -387,54 +382,22 @@ diffList l r
         Algo.Diff.Both _ _        ->
             pure ignore
 
--- Helper function to check when a diff part is present on both sides
-isBoth :: Algo.Diff.Diff a -> Bool
-isBoth p
-  | Algo.Diff.Both _ _ <- p = True
-  | otherwise               = False
-
 diffRecord
     :: (Eq a, Pretty a)
-    => Map Text (Expr s a) -> Map Text (Expr s a) -> Diff
+    => Map Text (Expr Void a) -> Map Text (Expr Void a) -> Diff
 diffRecord kvsL kvsR = braced (diffKeyVals colon kvsL kvsR)
 
 diffRecordLit
     :: (Eq a, Pretty a)
-    => Map Text (Expr s a) -> Map Text (Expr s a) -> Diff
+    => Map Text (Expr Void a) -> Map Text (Expr Void a) -> Diff
 diffRecordLit kvsL kvsR = braced (diffKeyVals equals kvsL kvsR)
 
 diffUnion
     :: (Eq a, Pretty a)
-    => Map Text (Expr s a) -> Map Text (Expr s a) -> Diff
-diffUnion kvsL kvsR = angled (diffKeyVals colon kvsL kvsR)
-
-diffUnionLit
-    :: (Eq a, Pretty a)
-    => Text
-    -> Text
-    -> Expr s a
-    -> Expr s a
-    -> Map Text (Expr s a)
-    -> Map Text (Expr s a)
-    -> Diff
-diffUnionLit kL kR vL vR kvsL kvsR =
-        langle
-    <>  " "
-    <>  format " " (diffLabel kL kR)
-    <>  equals
-    <>  " "
-    <>  format " " (diffExpression vL vR)
-    <>  halfAngled (diffKeyVals equals kvsL kvsR)
+    => Map Text (Maybe (Expr Void a)) -> Map Text (Maybe (Expr Void a)) -> Diff
+diffUnion kvsL kvsR = angled (diffKeysWith colon diffVals kvsL kvsR)
   where
-    halfAngled = enclosed (pipe <> " ") (pipe <> " ") rangle
-
-listSkeleton :: Diff
-listSkeleton =
-        lbracket
-    <>  " "
-    <>  ignore
-    <>  " "
-    <>  rbracket
+    diffVals = diffMaybe (colon <> " ") diff
 
 textSkeleton :: Diff
 textSkeleton =
@@ -552,32 +515,26 @@ skeleton (TextAppend {}) =
     <>  operator "++"
     <>  " "
     <>  ignore
-skeleton (ListLit {}) =
-        listSkeleton
-    <>  " "
-    <>  colon
-    <>  " "
-    <>  builtin "List"
-    <>  " "
-    <>  ignore
+skeleton (ListLit _ elems)
+    | null elems =
+            lbracket
+        <>  rbracket
+        <>  " "
+        <>  colon
+        <>  " "
+        <>  ignore
+    | otherwise =
+            lbracket
+        <>  " "
+        <>  ignore
+        <>  " "
+        <>  rbracket
 skeleton (ListAppend {}) =
         ignore
     <>  " "
     <>  operator "#"
     <>  " "
     <>  ignore
-skeleton (OptionalLit {}) =
-        lbracket
-    <>  " "
-    <>  ignore
-    <>  " "
-    <>  rbracket
-    <>  " "
-    <>  colon
-    <>  " "
-    <>  builtin "Optional"
-    <>  " "
-    <>  ignore
 skeleton (Record {}) =
         lbrace
     <>  " "
@@ -608,16 +565,6 @@ skeleton (Union {}) =
     <>  ignore
     <>  " "
     <>  rangle
-skeleton (UnionLit {}) =
-        langle
-    <>  " "
-    <>  ignore
-    <>  " "
-    <>  equals
-    <>  " "
-    <>  ignore
-    <>  " "
-    <>  rangle
 skeleton (Combine {}) =
         ignore
     <>  " "
@@ -642,8 +589,8 @@ skeleton (Merge {}) =
     <>  ignore
     <>  " "
     <>  ignore
-skeleton (Constructors {}) =
-        keyword "constructors"
+skeleton (ToMap {}) =
+        keyword "toMap"
     <>  " "
     <>  ignore
 skeleton (Field {}) =
@@ -663,8 +610,9 @@ skeleton x = token (Pretty.pretty x)
 mismatch :: Pretty a => Expr s a -> Expr s a -> Diff
 mismatch l r = difference (skeleton l) (skeleton r)
 
-diffExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
-diffExpression l@(Lam {}) r@(Lam {}) =
+-- | Render the difference between two expressions
+diff :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
+diff l@(Lam {}) r@(Lam {}) =
     enclosed' "  " (rarrow <> " ") (docs l r)
   where
     docs (Lam aL bL cL) (Lam aR bR cR) =
@@ -675,16 +623,16 @@ diffExpression l@(Lam {}) r@(Lam {}) =
             <>  format " " (diffLabel aL aR)
             <>  colon
             <>  " "
-            <>  format mempty (diffExpression bL bR)
+            <>  format mempty (diff bL bR)
             <>  rparen
 
     docs aL aR =
-        pure (diffExpression aL aR)
-diffExpression l@(Lam {}) r =
+        pure (diff aL aR)
+diff l@(Lam {}) r =
     mismatch l r
-diffExpression l r@(Lam {}) =
+diff l r@(Lam {}) =
     mismatch l r
-diffExpression l@(BoolIf {}) r@(BoolIf {}) =
+diff l@(BoolIf {}) r@(BoolIf {}) =
     enclosed' "      " (keyword "else" <> "  ") (docs l r)
   where
     docs (BoolIf aL bL cL) (BoolIf aR bR cR) =
@@ -692,64 +640,77 @@ diffExpression l@(BoolIf {}) r@(BoolIf {}) =
       where
         doc =   keyword "if"
             <>  " "
-            <>  format " " (diffExpression aL aR)
+            <>  format " " (diff aL aR)
             <>  keyword "then"
             <>  " "
-            <>  diffExpression bL bR
+            <>  diff bL bR
     docs aL aR =
-        pure (diffExpression aL aR)
-diffExpression l@(BoolIf {}) r =
+        pure (diff aL aR)
+diff l@(BoolIf {}) r =
     mismatch l r
-diffExpression l r@(BoolIf {}) =
+diff l r@(BoolIf {}) =
     mismatch l r
-diffExpression (Let asL bL ) (Let asR bR) =
-    enclosed' "" (keyword "in" <> "  ")
-        (Data.List.NonEmpty.zipWith docA asL asR <> pure docB)
+diff l@(Let {}) r@(Let {}) =
+    enclosed' "    " (keyword "in" <> "  ") (docs l r)
   where
-    docA (Binding cL dL eL) (Binding cR dR eR) = align doc
+    docs (Let (Binding _ aL _ bL _ cL) dL) (Let (Binding _ aR _ bR _ cR) dR) =
+        Data.List.NonEmpty.cons (align doc) (docs dL dR)
       where
+        bL' = fmap snd bL
+        bR' = fmap snd bR
+
         doc =   keyword "let"
             <>  " "
-            <>  format " " (diffLabel cL cR)
-            <>  format " " (diffMaybe (colon <> " ") diffExpression dL dR)
+            <>  format " " (diffLabel aL aR)
+            <>  format " " (diffMaybe (colon <> " ") diff bL' bR')
             <>  equals
             <>  " "
-            <>  diffExpression eL eR
-
-    docB = diffExpression bL bR
-diffExpression l@(Let {}) r =
+            <>  diff cL cR
+    docs aL aR = pure (diff aL aR)
+diff l@(Let {}) r =
     mismatch l r
-diffExpression l r@(Let {}) =
+diff l r@(Let {}) =
     mismatch l r
-diffExpression l@(Pi {}) r@(Pi {}) =
+diff l@(Pi {}) r@(Pi {}) =
     enclosed' "  " (rarrow <> " ") (docs l r)
   where
     docs (Pi aL bL cL) (Pi aR bR cR) =
         Data.List.NonEmpty.cons (align doc) (docs cL cR)
       where
         doc | same docA && same docB = ignore
+            | same docA =
+                format mempty docB
             | otherwise =
-                forall
-            <>  lparen
-            <>  format " " docA
-            <>  colon
-            <>  " "
-            <>  format mempty docB
-            <>  rparen
+                    forall
+                <>  lparen
+                <>  format " " docA
+                <>  colon
+                <>  " "
+                <>  format mempty docB
+                <>  rparen
           where
             docA = diffLabel aL aR
 
-            docB = diffExpression bL bR
+            docB = diff bL bR
 
-    docs aL aR = pure (diffExpression aL aR)
-diffExpression l@(Pi {}) r =
+    docs aL aR = pure (diff aL aR)
+diff l@(Pi {}) r =
     mismatch l r
-diffExpression l r@(Pi {}) =
+diff l r@(Pi {}) =
     mismatch l r
-diffExpression l r =
+diff (Assert aL) (Assert aR) =
+    align
+        (  "  " <> keyword "assert"
+        <> hardline <> colon <> " " <> diff aL aR
+        )
+diff l@(Assert {}) r =
+    mismatch l r
+diff l r@(Assert {}) =
+    mismatch l r
+diff l r =
     diffAnnotatedExpression l r
 
-diffAnnotatedExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffAnnotatedExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffAnnotatedExpression (Merge aL bL cL) (Merge aR bR cR) = align doc
   where
     doc =   keyword "merge"
@@ -761,29 +722,24 @@ diffAnnotatedExpression l@(Merge {}) r =
     mismatch l r
 diffAnnotatedExpression l r@(Merge {}) =
     mismatch l r
+diffAnnotatedExpression (ToMap aL bL) (ToMap aR bR) = align doc
+  where
+    doc =   keyword "toMap"
+        <>  " "
+        <>  format " " (diffImportExpression aL aR)
+        <>  diffMaybe (colon <> " ") diffApplicationExpression bL bR
+diffAnnotatedExpression l@(ToMap {}) r =
+    mismatch l r
+diffAnnotatedExpression l r@(ToMap {}) =
+    mismatch l r
 diffAnnotatedExpression (ListLit aL@(Just _) bL) (ListLit aR bR) = align doc
   where
     doc =   format " " (diffList bL bR)
-        <>  format " " (diffMaybe (colon <> " ") (diffApplicationExpression `on` App List) aL aR)
+        <>  format " " (diffMaybe (colon <> " ") diffApplicationExpression aL aR)
 diffAnnotatedExpression (ListLit aL bL) (ListLit aR@(Just _) bR) = align doc
   where
     doc =   format " " (diffList bL bR)
-        <>  format " " (diffMaybe (colon <> " ") (diffApplicationExpression `on` App List) aL aR)
-diffAnnotatedExpression (OptionalLit aL bL) (OptionalLit aR bR) =
-    align doc
-  where
-    doc =   lbracket
-        <>  " "
-        <>  format " " (diffMaybe mempty diffExpression bL bR)
-        <>  rbracket
-        <>  " "
-        <>  colon
-        <>  " "
-        <>  diffApplicationExpression (App Optional aL) (App Optional aR)
-diffAnnotatedExpression l@(OptionalLit {}) r =
-    mismatch l r
-diffAnnotatedExpression l r@(OptionalLit {}) =
-    mismatch l r
+        <>  format " " (diffMaybe (colon <> " ") diffApplicationExpression aL aR)
 diffAnnotatedExpression l@(Annot {}) r@(Annot {}) =
     enclosed' "  " (colon <> " ") (docs l r)
   where
@@ -792,7 +748,7 @@ diffAnnotatedExpression l@(Annot {}) r@(Annot {}) =
       where
         doc = diffOperatorExpression aL aR
     docs aL aR =
-        diffExpression aL aR :| []
+        diff aL aR :| []
 diffAnnotatedExpression l@(Annot {}) r =
     mismatch l r
 diffAnnotatedExpression l r@(Annot {}) =
@@ -800,17 +756,47 @@ diffAnnotatedExpression l r@(Annot {}) =
 diffAnnotatedExpression l r =
     diffOperatorExpression l r
 
-diffOperatorExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
-diffOperatorExpression = diffOrExpression
+{- Whitespace in diffs of operator expressions:
 
-diffOrExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+All indentation (whether pretty-printing or diffing) is a multiple of two
+spaces, so if the operator is one character long (like ?) then the diff pads
+the left margin to two space:
+
+    ␣␣e₀
+    ?␣e₁
+
+... but if the operator is two characters long (like ||) then the diff pads
+the left margin to four spaces:
+
+     ␣␣␣␣e₀
+     ||␣␣e₁
+-}
+diffOperatorExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
+diffOperatorExpression = diffImportAltExpression
+
+diffImportAltExpression :: (Pretty a, Eq a) => Expr Void a -> Expr Void a -> Diff
+diffImportAltExpression l@(ImportAlt {}) r@(ImportAlt {}) =
+    enclosed' "  " (operator "?" <> " ") (docs l r)
+  where
+    docs (ImportAlt aL bL) (ImportAlt aR bR) =
+        Data.List.NonEmpty.cons (diffOrExpression aL aR) (docs bL bR)
+    docs aL aR =
+        pure (diffOrExpression aL aR)
+diffImportAltExpression l@(ImportAlt {}) r =
+    mismatch l r
+diffImportAltExpression l r@(ImportAlt {}) =
+    mismatch l r
+diffImportAltExpression l r =
+    diffOrExpression l r
+
+diffOrExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffOrExpression l@(BoolOr {}) r@(BoolOr {}) =
     enclosed' "    " (operator "||" <> "  ") (docs l r)
   where
     docs (BoolOr aL bL) (BoolOr aR bR) =
-        Data.List.NonEmpty.cons (diffTextAppendExpression aL aR) (docs bL bR)
+        Data.List.NonEmpty.cons (diffPlusExpression aL aR) (docs bL bR)
     docs aL aR =
-        pure (diffTextAppendExpression aL aR)
+        pure (diffPlusExpression aL aR)
 diffOrExpression l@(BoolOr {}) r =
     mismatch l r
 diffOrExpression l r@(BoolOr {}) =
@@ -818,14 +804,14 @@ diffOrExpression l r@(BoolOr {}) =
 diffOrExpression l r =
     diffPlusExpression l r
 
-diffPlusExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffPlusExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffPlusExpression l@(NaturalPlus {}) r@(NaturalPlus {}) =
     enclosed' "  " (operator "+" <> " ") (docs l r)
   where
     docs (NaturalPlus aL bL) (NaturalPlus aR bR) =
-        Data.List.NonEmpty.cons (diffListAppendExpression aL aR) (docs bL bR)
+        Data.List.NonEmpty.cons (diffTextAppendExpression aL aR) (docs bL bR)
     docs aL aR =
-        pure (diffListAppendExpression aL aR)
+        pure (diffTextAppendExpression aL aR)
 diffPlusExpression l@(NaturalPlus {}) r =
     mismatch l r
 diffPlusExpression l r@(NaturalPlus {}) =
@@ -833,14 +819,14 @@ diffPlusExpression l r@(NaturalPlus {}) =
 diffPlusExpression l r =
     diffTextAppendExpression l r
 
-diffTextAppendExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffTextAppendExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffTextAppendExpression l@(TextAppend {}) r@(TextAppend {}) =
     enclosed' "    " (operator "++" <> "  ") (docs l r)
   where
     docs (TextAppend aL bL) (TextAppend aR bR) =
-        Data.List.NonEmpty.cons (diffPlusExpression aL aR) (docs bL bR)
+        Data.List.NonEmpty.cons (diffListAppendExpression aL aR) (docs bL bR)
     docs aL aR =
-        pure (diffPlusExpression aL aR)
+        pure (diffListAppendExpression aL aR)
 diffTextAppendExpression l@(TextAppend {}) r =
     mismatch l r
 diffTextAppendExpression l r@(TextAppend {}) =
@@ -848,7 +834,7 @@ diffTextAppendExpression l r@(TextAppend {}) =
 diffTextAppendExpression l r =
     diffListAppendExpression l r
 
-diffListAppendExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffListAppendExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffListAppendExpression l@(ListAppend {}) r@(ListAppend {}) =
     enclosed' "  " (operator "#" <> " ") (docs l r)
   where
@@ -863,7 +849,7 @@ diffListAppendExpression l r@(ListAppend {}) =
 diffListAppendExpression l r =
     diffAndExpression l r
 
-diffAndExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffAndExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffAndExpression l@(BoolAnd {}) r@(BoolAnd {}) =
     enclosed' "    " (operator "&&" <> "  ") (docs l r)
   where
@@ -878,7 +864,7 @@ diffAndExpression l r@(BoolAnd {}) =
 diffAndExpression l r =
     diffCombineExpression l r
 
-diffCombineExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffCombineExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffCombineExpression l@(Combine {}) r@(Combine {}) =
     enclosed' "  " (operator "∧" <> " ") (docs l r)
   where
@@ -893,7 +879,7 @@ diffCombineExpression l r@(Combine {}) =
 diffCombineExpression l r =
     diffPreferExpression l r
 
-diffPreferExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffPreferExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffPreferExpression l@(Prefer {}) r@(Prefer {}) =
     enclosed' "  " (operator "⫽" <> " ") (docs l r)
   where
@@ -908,7 +894,7 @@ diffPreferExpression l r@(Prefer {}) =
 diffPreferExpression l r =
     diffCombineTypesExpression l r
 
-diffCombineTypesExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffCombineTypesExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffCombineTypesExpression l@(CombineTypes {}) r@(CombineTypes {}) =
     enclosed' "  " (operator "*" <> " ") (docs l r)
   where
@@ -923,7 +909,7 @@ diffCombineTypesExpression l r@(CombineTypes {}) =
 diffCombineTypesExpression l r =
     diffTimesExpression l r
 
-diffTimesExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffTimesExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffTimesExpression l@(NaturalTimes {}) r@(NaturalTimes {}) =
     enclosed' "  " (operator "*" <> " ") (docs l r)
   where
@@ -938,7 +924,7 @@ diffTimesExpression l r@(NaturalTimes {}) =
 diffTimesExpression l r =
     diffEqualExpression l r
 
-diffEqualExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffEqualExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffEqualExpression l@(BoolEQ {}) r@(BoolEQ {}) =
     enclosed' "    " (operator "==" <> "  ") (docs l r)
   where
@@ -953,33 +939,43 @@ diffEqualExpression l r@(BoolEQ {}) =
 diffEqualExpression l r =
     diffNotEqualExpression l r
 
-diffNotEqualExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffNotEqualExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffNotEqualExpression l@(BoolNE {}) r@(BoolNE {}) =
     enclosed' "    " (operator "!=" <> "  ") (docs l r)
   where
     docs (BoolNE aL bL) (BoolNE aR bR) =
-        Data.List.NonEmpty.cons (diffApplicationExpression aL aR) (docs bL bR)
+        Data.List.NonEmpty.cons (diffEquivalentExpression aL aR) (docs bL bR)
     docs aL aR =
-        pure (diffApplicationExpression aL aR)
+        pure (diffEquivalentExpression aL aR)
 diffNotEqualExpression l@(BoolNE {}) r =
     mismatch l r
 diffNotEqualExpression l r@(BoolNE {}) =
     mismatch l r
 diffNotEqualExpression l r =
+    diffEquivalentExpression l r
+
+diffEquivalentExpression
+    :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
+diffEquivalentExpression l@(Equivalent {}) r@(Equivalent {}) =
+    enclosed' "  " (operator "≡" <> " ") (docs l r)
+  where
+    docs (Equivalent aL bL) (Equivalent aR bR) =
+        Data.List.NonEmpty.cons (diffApplicationExpression aL aR) (docs bL bR)
+    docs aL aR =
+        pure (diffApplicationExpression aL aR)
+diffEquivalentExpression l@(Equivalent {}) r =
+    mismatch l r
+diffEquivalentExpression l r@(Equivalent {}) =
+    mismatch l r
+diffEquivalentExpression l r =
     diffApplicationExpression l r
 
-diffApplicationExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffApplicationExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffApplicationExpression l@(App {}) r@(App {}) =
     enclosed' mempty mempty (Data.List.NonEmpty.reverse (docs l r))
   where
     docs (App aL bL) (App aR bR) =
         Data.List.NonEmpty.cons (diffImportExpression bL bR) (docs aL aR)
-    docs (Constructors aL) (Constructors aR) =
-        diffImportExpression aL aR :| [ keyword "constructors" ]
-    docs aL aR@(Constructors {}) =
-        pure (mismatch aL aR)
-    docs aL@(Constructors {}) aR =
-        pure (mismatch aL aR)
     docs (Some aL) (Some aR) =
         diffImportExpression aL aR :| [ builtin "Some" ]
     docs aL aR@(Some {}) =
@@ -992,12 +988,6 @@ diffApplicationExpression l@(App {}) r =
     mismatch l r
 diffApplicationExpression l r@(App {}) =
     mismatch l r
-diffApplicationExpression (Constructors l) (Constructors r) =
-    enclosed' mempty mempty (keyword "constructors" :| [ diffImportExpression l r ])
-diffApplicationExpression l@(Constructors {}) r =
-    mismatch l r
-diffApplicationExpression l r@(Constructors {}) =
-    mismatch l r
 diffApplicationExpression (Some l) (Some r) =
     enclosed' mempty mempty (builtin "Some" :| [ diffImportExpression l r ])
 diffApplicationExpression l@(Some {}) r =
@@ -1007,7 +997,7 @@ diffApplicationExpression l r@(Some {}) =
 diffApplicationExpression l r =
     diffImportExpression l r
 
-diffImportExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffImportExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffImportExpression (Embed l) (Embed r) =
     diffPretty l r
 diffImportExpression l@(Embed {}) r =
@@ -1017,14 +1007,16 @@ diffImportExpression l r@(Embed {}) =
 diffImportExpression l r =
     diffSelectorExpression l r
 
-diffSelectorExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffSelectorExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffSelectorExpression l@(Field {}) r@(Field {}) =
     enclosed' "  " (dot <> " ") (Data.List.NonEmpty.reverse (docs l r))
   where
     docs (Field aL bL) (Field aR bR) =
         Data.List.NonEmpty.cons (diffLabel bL bR) (docs aL aR)
-    docs (Project aL bL) (Project aR bR) =
+    docs (Project aL (Left bL)) (Project aR (Left bR)) =
         Data.List.NonEmpty.cons (diffLabels bL bR) (docs aL aR)
+    docs (Project aL (Right bL)) (Project aR (Right bR)) =
+        Data.List.NonEmpty.cons (diff bL bR) (docs aL aR)
     docs aL aR =
         pure (diffPrimitiveExpression aL aR)
 diffSelectorExpression l@(Field {}) r =
@@ -1036,8 +1028,10 @@ diffSelectorExpression l@(Project {}) r@(Project {}) =
   where
     docs (Field aL bL) (Field aR bR) =
         Data.List.NonEmpty.cons (diffLabel bL bR) (docs aL aR)
-    docs (Project aL bL) (Project aR bR) =
+    docs (Project aL (Left bL)) (Project aR (Left bR)) =
         Data.List.NonEmpty.cons (diffLabels bL bR) (docs aL aR)
+    docs (Project aL (Right bL)) (Project aR (Right bR)) =
+        Data.List.NonEmpty.cons (diff bL bR) (docs aL aR)
     docs aL aR =
         pure (diffPrimitiveExpression aL aR)
 diffSelectorExpression l@(Project {}) r =
@@ -1047,7 +1041,7 @@ diffSelectorExpression l r@(Project {}) =
 diffSelectorExpression l r =
     diffPrimitiveExpression l r
 
-diffPrimitiveExpression :: (Eq a, Pretty a) => Expr s a -> Expr s a -> Diff
+diffPrimitiveExpression :: (Eq a, Pretty a) => Expr Void a -> Expr Void a -> Diff
 diffPrimitiveExpression (Var aL) (Var aR) =
     diffVar aL aR
 diffPrimitiveExpression l@(Var {}) r =
@@ -1114,6 +1108,12 @@ diffPrimitiveExpression l@NaturalShow r =
     mismatch l r
 diffPrimitiveExpression l r@NaturalShow =
     mismatch l r
+diffPrimitiveExpression NaturalSubtract NaturalSubtract =
+    "…"
+diffPrimitiveExpression l@NaturalSubtract r =
+    mismatch l r
+diffPrimitiveExpression l r@NaturalSubtract =
+    mismatch l r
 diffPrimitiveExpression Integer Integer =
     "…"
 diffPrimitiveExpression l@Integer r =
@@ -1150,6 +1150,12 @@ diffPrimitiveExpression l@Text r =
     mismatch l r
 diffPrimitiveExpression l r@Text =
     mismatch l r
+diffPrimitiveExpression TextShow TextShow =
+    "…"
+diffPrimitiveExpression l@TextShow r =
+    mismatch l r
+diffPrimitiveExpression l r@TextShow =
+    mismatch l r
 diffPrimitiveExpression List List =
     "…"
 diffPrimitiveExpression l@List r =
@@ -1159,10 +1165,6 @@ diffPrimitiveExpression l r@List =
 diffPrimitiveExpression (ListLit Nothing bL) (ListLit Nothing bR) = align doc
   where
     doc = format " " (diffList bL bR)
-diffPrimitiveExpression l@(ListLit {}) r =
-    mismatch l r
-diffPrimitiveExpression l r@(ListLit {}) =
-    mismatch l r
 diffPrimitiveExpression ListBuild ListBuild =
     "…"
 diffPrimitiveExpression l@ListBuild r =
@@ -1277,15 +1279,9 @@ diffPrimitiveExpression l@(Union {}) r =
     mismatch l r
 diffPrimitiveExpression l r@(Union {}) =
     mismatch l r
-diffPrimitiveExpression (UnionLit aL bL cL) (UnionLit aR bR cR) =
-    diffUnionLit aL aR bL bR cL cR
-diffPrimitiveExpression l@(UnionLit {}) r =
-    mismatch l r
-diffPrimitiveExpression l r@(UnionLit {}) =
-    mismatch l r
 diffPrimitiveExpression aL aR =
     if same doc
     then ignore
     else align ("( " <> doc <> hardline <> ")")
   where
-    doc = diffExpression aL aR
+    doc = diff aL aR
