@@ -7,29 +7,23 @@ module Dhall.Hash
       hash
     ) where
 
-import Dhall.Binary (StandardVersion)
-import Dhall.Parser (exprFromText)
-import Dhall.Import (hashExpressionToCode, standardVersion)
-import Lens.Family (set)
+import Dhall.Import (hashExpressionToCode)
+import Dhall.Util (Censor, Input(..))
 
 import qualified Control.Monad.Trans.State.Strict as State
 import qualified Control.Exception
 import qualified Dhall.Core
 import qualified Dhall.Import
 import qualified Dhall.TypeCheck
+import qualified Dhall.Util
 import qualified Data.Text.IO
 
 -- | Implementation of the @dhall hash@ subcommand
-hash :: StandardVersion -> IO ()
-hash _standardVersion = do
-    inText <- Data.Text.IO.getContents
+hash :: Censor -> IO ()
+hash censor = do
+    parsedExpression <- Dhall.Util.getExpression censor StandardInput
 
-    parsedExpression <- case exprFromText "(stdin)" inText of
-        Left  exception        -> Control.Exception.throwIO exception
-        Right parsedExpression -> return parsedExpression
-
-    let status =
-            set standardVersion _standardVersion (Dhall.Import.emptyStatus ".")
+    let status = Dhall.Import.emptyStatus "."
 
     resolvedExpression <- State.evalStateT (Dhall.Import.loadWith parsedExpression) status
 
@@ -40,5 +34,4 @@ hash _standardVersion = do
     let normalizedExpression =
             Dhall.Core.alphaNormalize (Dhall.Core.normalize resolvedExpression)
 
-    Data.Text.IO.putStrLn
-        (hashExpressionToCode _standardVersion normalizedExpression)
+    Data.Text.IO.putStrLn (hashExpressionToCode normalizedExpression)
