@@ -1,4 +1,5 @@
-{-# LANGUAGE RecordWildCards            #-}
+{-# LANGUAGE CPP             #-}
+{-# LANGUAGE RecordWildCards #-}
 
 -- | This module contains Dhall's parsing logic
 
@@ -6,12 +7,14 @@ module Dhall.Parser (
     -- * Utilities
       exprFromText
     , exprAndHeaderFromText
+    , censor
 
     -- * Parsers
     , expr, exprA
 
     -- * Types
     , Src(..)
+    , SourcedException(..)
     , ParseError(..)
     , Parser(..)
     ) where
@@ -21,9 +24,13 @@ import Data.Semigroup (Semigroup(..))
 import Data.Text (Text)
 import Data.Void (Void)
 import Dhall.Core
+import Dhall.Src (Src(..))
 import Prelude hiding (const, pi)
+import Text.Megaparsec (ParseErrorBundle(..), PosState(..))
 
+import qualified Data.Char
 import qualified Data.Text
+import qualified Dhall.Core      as Core
 import qualified Text.Megaparsec
 
 import Dhall.Parser.Combinators
@@ -32,7 +39,7 @@ import Dhall.Parser.Expression
 
 -- | Parser for a top-level Dhall expression
 expr :: Parser (Expr Src Import)
-expr = exprA import_
+expr = exprA (Text.Megaparsec.try import_)
 
 -- | Parser for a top-level Dhall expression. The expression is parameterized
 -- over any parseable type, allowing the language to be extended as needed.
@@ -40,14 +47,40 @@ exprA :: Parser a -> Parser (Expr Src a)
 exprA = completeExpression
 
 -- | A parsing error
-data ParseError = ParseError
-    { unwrap :: Text.Megaparsec.ParseErrorBundle Text Void
+data ParseError = ParseError {
+#if MIN_VERSION_megaparsec(7, 0, 0)
+      unwrap :: Text.Megaparsec.ParseErrorBundle Text Void
+#else
+      unwrap :: Text.Megaparsec.ParseError Char Void
+#endif
     , input  :: Text
     }
 
+{-| Replace the source code with spaces when rendering error messages
+
+    This utility is used to implement the @--censor@ flag
+-}
+censor :: ParseError -> ParseError
+censor parseError =
+    parseError
+        { unwrap =
+            (unwrap parseError)
+                { bundlePosState =
+                    (bundlePosState (unwrap parseError))
+                        { pstateInput =
+                            Core.censorText
+                                (pstateInput (bundlePosState (unwrap parseError)))
+                        }
+                }
+        }
+
 instance Show ParseError where
     show (ParseError {..}) =
+#if MIN_VERSION_megaparsec(7, 0, 0)
       "\n\ESC[1;31mError\ESC[0m: Invalid input\n\n" <> Text.Megaparsec.errorBundlePretty unwrap
+#else
+      "\n\ESC[1;31mError\ESC[0m: Invalid input\n\n" <> Text.Megaparsec.parseErrorPretty unwrap
+#endif
 
 instance Exception ParseError
 
@@ -78,7 +111,7 @@ exprAndHeaderFromText
     -> Either ParseError (Text, Expr Src Import)
 exprAndHeaderFromText delta text = case result of
     Left errInfo   -> Left (ParseError { unwrap = errInfo, input = text })
-    Right (txt, r) -> Right (Data.Text.dropWhileEnd (/= '\n') txt, r)
+    Right (txt, r) -> Right (stripHeader txt, r)
   where
     parser = do
         (bytes, _) <- Text.Megaparsec.match whitespace
@@ -87,3 +120,5 @@ exprAndHeaderFromText delta text = case result of
         return (bytes, r)
 
     result = Text.Megaparsec.parse (unParser parser) delta text
+
+    stripHeader = Data.Text.dropWhile Data.Char.isSpace . Data.Text.dropWhileEnd (/= '\n')
