@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP                 #-}
 {-# LANGUAGE NamedFieldPuns      #-}
 {-# LANGUAGE RecordWildCards     #-}
 {-# LANGUAGE OverloadedLists     #-}
@@ -14,40 +15,92 @@ import Data.List.NonEmpty (NonEmpty(..))
 import Data.Semigroup (Semigroup(..))
 import Data.Text (Text)
 import Dhall.Core
+import Dhall.Src (Src(..))
 import Prelude hiding (const, pi)
 import Text.Parser.Combinators (choice, try, (<?>))
 
-import qualified Crypto.Hash
+import qualified Control.Monad
 import qualified Data.ByteArray.Encoding
 import qualified Data.ByteString
-import qualified Data.Char
+import qualified Data.Char               as Char
 import qualified Data.Foldable
+import qualified Data.List
 import qualified Data.List.NonEmpty
 import qualified Data.Sequence
 import qualified Data.Text
 import qualified Data.Text.Encoding
+import qualified Dhall.Crypto
 import qualified Text.Megaparsec
+#if !MIN_VERSION_megaparsec(7, 0, 0)
+import qualified Text.Megaparsec.Char    as Text.Megaparsec
+#endif
 import qualified Text.Parser.Char
 
 import Dhall.Parser.Combinators
 import Dhall.Parser.Token
 
+getSourcePos :: Text.Megaparsec.MonadParsec e s m =>
+                m Text.Megaparsec.SourcePos
+getSourcePos =
+#if MIN_VERSION_megaparsec(7, 0, 0)
+    Text.Megaparsec.getSourcePos
+#else
+    Text.Megaparsec.getPosition
+#endif
+{-# INLINE getSourcePos #-}
+
+getOffset :: Text.Megaparsec.MonadParsec e s m => m Int
+#if MIN_VERSION_megaparsec(7, 0, 0)
+getOffset = Text.Megaparsec.stateOffset <$> Text.Megaparsec.getParserState
+#else
+getOffset = Text.Megaparsec.stateTokensProcessed <$> Text.Megaparsec.getParserState
+#endif
+{-# INLINE getOffset #-}
+
+setOffset :: Text.Megaparsec.MonadParsec e s m => Int -> m ()
+#if MIN_VERSION_megaparsec(7, 0, 0)
+setOffset o = Text.Megaparsec.updateParserState $ \(Text.Megaparsec.State s _ pst) ->
+  Text.Megaparsec.State s o pst
+#else
+setOffset o = Text.Megaparsec.updateParserState $ \(Text.Megaparsec.State s p _ stw) ->
+  Text.Megaparsec.State s p o stw
+#endif
+{-# INLINE setOffset #-}
+
+src :: Parser a -> Parser Src
+src parser = do
+    before      <- getSourcePos
+    (tokens, _) <- Text.Megaparsec.match parser
+    after       <- getSourcePos
+    return (Src before after tokens)
+
 noted :: Parser (Expr Src a) -> Parser (Expr Src a)
 noted parser = do
-    before      <- Text.Megaparsec.getSourcePos
+    before      <- getSourcePos
     (tokens, e) <- Text.Megaparsec.match parser
-    after       <- Text.Megaparsec.getSourcePos
+    after       <- getSourcePos
     let src₀ = Src before after tokens
     case e of
         Note src₁ _ | laxSrcEq src₀ src₁ -> return e
         _                                -> return (Note src₀ e)
 
-shallowDenote :: Expr s a -> Expr s a
-shallowDenote (Note _ e) = shallowDenote e
-shallowDenote         e  = e
-
 completeExpression :: Parser a -> Parser (Expr Src a)
 completeExpression embedded = completeExpression_
+  where
+    Parsers {..} = parsers embedded
+
+importExpression :: Parser a -> Parser (Expr Src a)
+importExpression embedded = importExpression_
+  where
+    Parsers {..} = parsers embedded
+
+data Parsers a = Parsers
+    { completeExpression_ :: Parser (Expr Src a)
+    , importExpression_   :: Parser (Expr Src a)
+    }
+
+parsers :: Parser a -> Parsers a
+parsers embedded = Parsers {..}
   where
     completeExpression_ = do
         whitespace
@@ -61,6 +114,7 @@ completeExpression embedded = completeExpression_
                 , alternative2
                 , alternative3
                 , alternative4
+                , alternative5
                 ]
             ) <?> "expression"
       where
@@ -86,18 +140,29 @@ completeExpression embedded = completeExpression_
 
         alternative2 = do
             let binding = do
-                    _let
+                    _letOnly
+
+                    src0 <- src nonemptyWhitespace
+
+                    c <- labelOnly
+
+                    src1 <- src whitespace
 
-                    c <- label
                     d <- optional (do
-                        _colon
-                        expression )
+                        _colonOnly
 
-                    _equal
+                        src2 <- src nonemptyWhitespace
 
-                    e <- expression
+                        e <- expression
+                        return (Just src2, e) )
 
-                    return (Binding c d e)
+                    _equalOnly
+
+                    src3 <- src whitespace
+
+                    f <- expression
+
+                    return (Binding (Just src0) c (Just src1) d (Just src3) f)
 
             as <- Data.List.NonEmpty.some1 binding
 
@@ -105,7 +170,24 @@ completeExpression embedded = completeExpression_
 
             b <- expression
 
-            return (Let as b)
+            -- 'Note's in let-in-let:
+            --
+            -- Subsequent @let@s that are not separated by an @in@ only get a
+            -- single surrounding 'Note'. For example:
+            --
+            -- let x = a
+            -- let y = b
+            -- in  let z = c
+            --     in x
+            --
+            -- is parsed as
+            --
+            -- (Note …
+            --   (Let x …
+            --     (Let y …
+            --       (Note …
+            --         (Let z …
+            return (Dhall.Core.wrapInLets as b)
 
         alternative3 = do
             _forall
@@ -119,6 +201,12 @@ completeExpression embedded = completeExpression_
             return (Pi a b c)
 
         alternative4 = do
+            _assert
+            _colon
+            a <- expression
+            return (Assert a)
+
+        alternative5 = do
             a <- operatorExpression
 
             let alternative4A = do
@@ -131,19 +219,13 @@ completeExpression embedded = completeExpression_
 
                     b <- expression
 
-                    case (shallowDenote a, shallowDenote b) of
-                        (ListLit _ xs, App f c) ->
-                            case shallowDenote f of
-                                List     ->
-                                    return (ListLit (Just c) xs)
-                                Optional -> case xs of
-                                    [x] -> return (OptionalLit c (Just x))
-                                    []  -> return (OptionalLit c Nothing)
-                                    _   -> return (Annot a b)
-                                _ ->
-                                    return (Annot a b)
-                        (Merge c d _, e) ->
-                            return (Merge c d (Just e))
+                    case shallowDenote a of
+                        ListLit _ [] ->
+                            return (ListLit (Just b) [])
+                        Merge c d _ ->
+                            return (Merge c d (Just b))
+                        ToMap c _ ->
+                            return (ToMap c (Just b))
                         _ -> return (Annot a b)
 
             alternative4A <|> alternative4B <|> pure a
@@ -175,9 +257,11 @@ completeExpression embedded = completeExpression_
                 CombineTypes <$ _combineTypes
             <|> Prefer       <$ _prefer
             <|> NaturalTimes <$ _times
-            <|> BoolNE       <$ _notEqual
+            <|> BoolEQ       <$ _doubleEqual
 
-    precedence3Operator = BoolEQ <$ _doubleEqual
+    precedence3Operator
+            =   BoolNE     <$ _notEqual
+            <|> Equivalent <$ _equivalent
 
     precedence0Expression =
             makeOperatorExpression precedence1Expression precedence0Operator
@@ -192,11 +276,10 @@ completeExpression embedded = completeExpression_
             makeOperatorExpression applicationExpression precedence3Operator
 
     applicationExpression = do
-            f <-    (do _constructors; return Constructors)
-                <|> (do _Some; return Some)
+            f <-    (do _Some; return Some)
                 <|> return id
-            a <- noted importExpression
-            b <- Text.Megaparsec.many (noted importExpression)
+            a <- noted importExpression_
+            b <- Text.Megaparsec.many (noted importExpression_)
             return (foldl app (f a) b)
           where
             app nL@(Note (Src before _ bytesL) _) nR@(Note (Src _ after bytesR) _) =
@@ -204,7 +287,7 @@ completeExpression embedded = completeExpression_
             app nL nR =
                 App nL nR
 
-    importExpression = noted (choice [ alternative0, alternative1 ])
+    importExpression_ = noted (choice [ alternative0, alternative1 ])
           where
             alternative0 = do
                 a <- embedded
@@ -215,9 +298,18 @@ completeExpression embedded = completeExpression_
     selectorExpression = noted (do
             a <- primitiveExpression
 
-            let left  x  e = Field   e x
-            let right xs e = Project e xs
-            b <- Text.Megaparsec.many (try (do _dot; fmap left label <|> fmap right labels))
+            let recordType = _openParens *> expression <* _closeParens
+
+            let field               x  e = Field   e  x
+            let projectBySet        xs e = Project e (Left  xs)
+            let projectByExpression xs e = Project e (Right xs)
+
+            let alternatives =
+                        fmap field               anyLabel
+                    <|> fmap projectBySet        labels
+                    <|> fmap projectByExpression recordType
+
+            b <- Text.Megaparsec.many (try (do _dot; alternatives))
             return (foldl (\e k -> k e) a b) )
 
     primitiveExpression =
@@ -231,6 +323,7 @@ completeExpression embedded = completeExpression_
                     , alternative05
                     , alternative06
                     , alternative07
+                    , alternative08
                     , alternative37
                     , alternative09
 
@@ -240,10 +333,10 @@ completeExpression embedded = completeExpression_
             <|> alternative38
           where
             alternative00 = do
-                n <- Text.Megaparsec.getOffset
+                n <- getOffset
                 a <- try doubleLiteral
                 b <- if isInfinite a
-                       then Text.Megaparsec.setOffset n *> fail "double out of bounds"
+                       then setOffset n *> fail "double out of bounds"
                        else return a
                 return (DoubleLit b)
 
@@ -263,20 +356,21 @@ completeExpression embedded = completeExpression_
                 _closeBrace
                 return a ) <?> "record type or literal"
 
-            alternative05 = (do
-                _openAngle
-                a <- unionTypeOrLiteral
-                _closeAngle
-                return a ) <?> "union type or literal"
+            alternative05 = unionType
 
             alternative06 = listLiteral
 
             alternative07 = do
                 _merge
-                a <- importExpression
-                b <- importExpression
+                a <- importExpression_
+                b <- importExpression_ <?> "second argument to ❰merge❱"
                 return (Merge a b Nothing)
 
+            alternative08 = do
+                _toMap
+                a <- importExpression_
+                return (ToMap a Nothing)
+
             alternative09 = do
                 a <- try doubleInfinity
                 return (DoubleLit a)
@@ -306,7 +400,7 @@ completeExpression embedded = completeExpression_
                             , NaturalIsZero    <$ _NaturalIsZero
                             , NaturalEven      <$ _NaturalEven
                             , NaturalOdd       <$ _NaturalOdd
-                            , NaturalToInteger <$ _NaturalToInteger
+                            , NaturalSubtract  <$ _NaturalSubtract
                             , NaturalToInteger <$ _NaturalToInteger
                             , NaturalShow      <$ _NaturalShow
                             , Natural          <$ _Natural
@@ -346,7 +440,8 @@ completeExpression embedded = completeExpression_
                     'S' ->    Const Sort       <$ _Sort
                     'T' ->
                         choice
-                            [ Text             <$ _Text
+                            [ TextShow         <$ _TextShow
+                            , Text             <$ _Text
                             , BoolLit True     <$ _True
                             , Const Type       <$ _Type
                             ]
@@ -389,7 +484,7 @@ completeExpression embedded = completeExpression_
                     ) && c /= '$'
 
             unescapedCharacterSlow = do
-                _ <- Text.Megaparsec.single '$'
+                _ <- Text.Parser.Char.char '$'
                 return (Chunks [] "$")
 
             escapedCharacter = do
@@ -428,12 +523,28 @@ completeExpression embedded = completeExpression_
 
                 unicode = do
                     _  <- Text.Parser.Char.char 'u';
-                    n0 <- hexNumber
-                    n1 <- hexNumber
-                    n2 <- hexNumber
-                    n3 <- hexNumber
-                    let n = ((n0 * 16 + n1) * 16 + n2) * 16 + n3
-                    return (Data.Char.chr n)
+
+                    let toNumber = Data.List.foldl' (\x y -> x * 16 + y) 0
+
+                    let fourCharacterEscapeSequence =
+                            fmap toNumber (Control.Monad.replicateM 4 hexNumber)
+
+                    let bracedEscapeSequence = do
+                            _  <- Text.Parser.Char.char '{'
+                            ns <- some hexNumber
+
+                            let number = toNumber ns
+
+                            Control.Monad.guard (number <= 0x10FFFF && validCodepoint (Char.chr number))
+                                <|> fail "Invalid Unicode code point"
+
+                            _  <- Text.Parser.Char.char '}'
+
+                            return (toNumber ns)
+
+                    n <- bracedEscapeSequence <|> fourCharacterEscapeSequence
+
+                    return (Char.chr n)
 
     doubleQuotedLiteral = do
             _      <- Text.Parser.Char.char '"'
@@ -531,14 +642,14 @@ completeExpression embedded = completeExpression_
             alternative2 = return (Record mempty)
 
     nonEmptyRecordTypeOrLiteral = do
-            a <- label
+            a <- anyLabel
 
             let nonEmptyRecordType = do
                     _colon
                     b <- expression
                     e <- Text.Megaparsec.many (do
                         _comma
-                        c <- label
+                        c <- anyLabel
                         _colon
                         d <- expression
                         return (c, d) )
@@ -550,7 +661,7 @@ completeExpression embedded = completeExpression_
                     b <- expression
                     e <- Text.Megaparsec.many (do
                         _comma
-                        c <- label
+                        c <- anyLabel
                         _equal
                         d <- expression
                         return (c, d) )
@@ -559,43 +670,21 @@ completeExpression embedded = completeExpression_
 
             nonEmptyRecordType <|> nonEmptyRecordLiteral
 
-    unionTypeOrLiteral =
-                nonEmptyUnionTypeOrLiteral
-            <|> return (Union mempty)
+    unionType = (do
+            _openAngle
+
+            let unionTypeEntry = do
+                    a <- anyLabel
+                    b <- optional (do _colon; expression)
+                    return (a, b)
+
+            kvs <- Text.Megaparsec.sepBy unionTypeEntry _bar
 
-    nonEmptyUnionTypeOrLiteral = do
-            (f, kvs) <- loop
             m <- toMap kvs
-            return (f m)
-          where
-            loop = do
-                a <- label
 
-                let alternative0 = do
-                        _equal
-                        b <- expression
-                        kvs <- Text.Megaparsec.many (do
-                            _bar
-                            c <- label
-                            _colon
-                            d <- expression
-                            return (c, d) )
-                        return (UnionLit a b, kvs)
+            _closeAngle
 
-                let alternative1 = do
-                        _colon
-                        b <- expression
-
-                        let alternative2 = do
-                                _bar
-                                (f, kvs) <- loop
-                                return (f, (a, b):kvs)
-
-                        let alternative3 = return (Union, [(a, b)])
-
-                        alternative2 <|> alternative3
-
-                alternative0 <|> alternative1
+            return (Union m) ) <?> "union type"
 
     listLiteral = (do
             _openBracket
@@ -629,24 +718,24 @@ localRaw =
   where
     parentPath = do
         _    <- ".." :: Parser Text
-        file <- file_
+        file <- file_ FileComponent
 
         return (Local Parent file)
 
     herePath = do
         _    <- "." :: Parser Text
-        file <- file_
+        file <- file_ FileComponent
 
         return (Local Here file)
 
     homePath = do
         _    <- "~" :: Parser Text
-        file <- file_
+        file <- file_ FileComponent
 
         return (Local Home file)
 
     absolutePath = do
-        file <- file_
+        file <- file_ FileComponent
 
         return (Local Absolute file)
 
@@ -662,7 +751,7 @@ http = do
     whitespace
     headers <- optional (do
         _using
-        (importHashed_ <|> (_openParens *> importHashed_ <* _closeParens)) )
+        importExpression import_ )
     return (Remote (url { headers }))
 
 missing :: Parser ImportType
@@ -679,23 +768,24 @@ importType_ = do
 
     choice [ local, http, env, missing ]
 
+importHash_ :: Parser Dhall.Crypto.SHA256Digest
+importHash_ = do
+    _ <- Text.Parser.Char.text "sha256:"
+    text <- count 64 (satisfy hexdig <?> "hex digit")
+    whitespace
+    let strictBytes16 = Data.Text.Encoding.encodeUtf8 text
+    strictBytes <- case Data.ByteArray.Encoding.convertFromBase Base16 strictBytes16 of
+        Left  string      -> fail string
+        Right strictBytes -> return (strictBytes :: Data.ByteString.ByteString)
+    case Dhall.Crypto.sha256DigestFromByteString strictBytes of
+      Nothing -> fail "Invalid sha256 hash"
+      Just h  -> pure h
+
 importHashed_ :: Parser ImportHashed
 importHashed_ = do
     importType <- importType_
     hash       <- optional importHash_
     return (ImportHashed {..})
-  where
-    importHash_ = do
-        _ <- Text.Parser.Char.text "sha256:"
-        text <- count 64 (satisfy hexdig <?> "hex digit")
-        whitespace
-        let strictBytes16 = Data.Text.Encoding.encodeUtf8 text
-        strictBytes <- case Data.ByteArray.Encoding.convertFromBase Base16 strictBytes16 of
-            Left  string      -> fail string
-            Right strictBytes -> return (strictBytes :: Data.ByteString.ByteString)
-        case Crypto.Hash.digestFromByteString strictBytes of
-          Nothing -> fail "Invalid sha256 hash"
-          Just h  -> pure h
 
 import_ :: Parser Import
 import_ = (do
@@ -704,9 +794,8 @@ import_ = (do
     return (Import {..}) ) <?> "import"
   where
     alternative = do
-        _as
-        _Text
-        return RawText
+      _as
+      (_Text >> pure RawText) <|> (_Location >> pure Location)
 
 -- | Similar to `Dhall.Core.renderChunks` except that this doesn't bother to
 -- render interpolated expressions to avoid a `Buildable a` constraint.  The
@@ -747,10 +836,16 @@ unlinesLiteral :: NonEmpty (Chunks s a) -> Chunks s a
 unlinesLiteral chunks =
     Data.Foldable.fold (Data.List.NonEmpty.intersperse "\n" chunks)
 
-leadingSpaces :: Chunks s a -> Int
-leadingSpaces chunks =
-    Data.Text.length (Data.Text.takeWhile Data.Char.isSpace firstText)
+emptyLine :: Chunks s a -> Bool
+emptyLine (Chunks [] ""  ) = True
+emptyLine (Chunks [] "\r") = True  -- So that `\r\n` is treated as a blank line
+emptyLine  _               = False
+
+leadingSpaces :: Chunks s a -> Text
+leadingSpaces chunks = Data.Text.takeWhile isSpace firstText
   where
+    isSpace c = c == '\x20' || c == '\x09'
+
     firstText =
         case chunks of
             Chunks                []  suffix -> suffix
@@ -768,6 +863,26 @@ toDoubleQuoted literal =
   where
     literals = linesLiteral literal
 
-    l :| ls = literals
+    sharedPrefix ab ac =
+        case Data.Text.commonPrefixes ab ac of
+            Just (a, _b, _c) -> a
+            Nothing          -> ""
 
-    indent = Data.Foldable.foldl' min (leadingSpaces l) (fmap leadingSpaces ls)
+    -- The standard specifies to filter out blank lines for all lines *except*
+    -- for the last line
+    filteredLines = newInit <> pure oldLast
+      where
+        oldInit = Data.List.NonEmpty.init literals
+
+        oldLast = Data.List.NonEmpty.last literals
+
+        newInit = filter (not . emptyLine) oldInit
+
+    longestSharedPrefix =
+        case filteredLines of
+            l : ls ->
+                Data.Foldable.foldl' sharedPrefix (leadingSpaces l) (fmap leadingSpaces ls)
+            [] ->
+                ""
+
+    indent = Data.Text.length longestSharedPrefix
