@@ -3,11 +3,16 @@
 -- | Parse Dhall tokens. Even though we don't have a tokenizer per-se this
 ---  module is useful for keeping some small parsing utilities.
 module Dhall.Parser.Token (
+    validCodepoint,
     whitespace,
+    nonemptyWhitespace,
     bashEnvironmentVariable,
     posixEnvironmentVariable,
+    ComponentType(..),
     file_,
+    labelOnly,
     label,
+    anyLabel,
     labels,
     httpRaw,
     hexdig,
@@ -21,12 +26,14 @@ module Dhall.Parser.Token (
     _if,
     _then,
     _else,
+    _letOnly,
     _let,
     _in,
     _as,
     _using,
     _merge,
-    _constructors,
+    _toMap,
+    _assert,
     _Some,
     _None,
     _NaturalFold,
@@ -36,6 +43,7 @@ module Dhall.Parser.Token (
     _NaturalOdd,
     _NaturalToInteger,
     _NaturalShow,
+    _NaturalSubtract,
     _IntegerShow,
     _IntegerToDouble,
     _DoubleShow,
@@ -53,6 +61,7 @@ module Dhall.Parser.Token (
     _Integer,
     _Double,
     _Text,
+    _TextShow,
     _List,
     _True,
     _False,
@@ -60,6 +69,8 @@ module Dhall.Parser.Token (
     _Type,
     _Kind,
     _Sort,
+    _Location,
+    _equalOnly,
     _equal,
     _or,
     _plus,
@@ -80,8 +91,10 @@ module Dhall.Parser.Token (
     _comma,
     _openParens,
     _closeParens,
+    _colonOnly,
     _colon,
     _at,
+    _equivalent,
     _missing,
     _importAlt,
     _combine,
@@ -104,12 +117,14 @@ import Prelude hiding (const, pi)
 import Text.Parser.Combinators (choice, try, (<?>))
 
 import qualified Control.Monad
-import qualified Data.Char
+import qualified Data.Char                  as Char
 import qualified Data.HashSet
 import qualified Data.List.NonEmpty
 import qualified Data.Text
 import qualified Dhall.Set
+import qualified Network.URI.Encode         as URI.Encode
 import qualified Text.Megaparsec
+import qualified Text.Megaparsec.Char.Lexer
 import qualified Text.Parser.Char
 import qualified Text.Parser.Combinators
 
@@ -118,6 +133,11 @@ import Prelude hiding (const, pi)
 
 import qualified Text.Parser.Token
 
+validCodepoint :: Char -> Bool
+validCodepoint c =
+    not (category == Char.Surrogate || category == Char.NotAssigned)
+  where
+    category = Char.generalCategory c
 
 whitespace :: Parser ()
 whitespace = Text.Parser.Combinators.skipMany whitespaceChunk
@@ -131,6 +151,9 @@ alpha c = ('\x41' <= c && c <= '\x5A') || ('\x61' <= c && c <= '\x7A')
 digit :: Char -> Bool
 digit c = '\x30' <= c && c <= '\x39'
 
+alphaNum :: Char -> Bool
+alphaNum c = alpha c || digit c
+
 hexdig :: Char -> Bool
 hexdig c =
         ('0' <= c && c <= '9')
@@ -159,13 +182,15 @@ doubleInfinity = (do
 integerLiteral :: Parser Integer
 integerLiteral = (do
     sign <- signPrefix
-    a <- Text.Parser.Token.natural
+    a <- Text.Megaparsec.Char.Lexer.decimal
+    whitespace
     return (sign a) ) <?> "integer literal"
 
 naturalLiteral :: Parser Natural
 naturalLiteral = (do
-    a <- Text.Parser.Token.natural
-    return (fromIntegral a) ) <?> "natural literal"
+    a <- Text.Megaparsec.Char.Lexer.decimal
+    whitespace
+    return a ) <?> "natural literal"
 
 identifier :: Parser Var
 identifier = do
@@ -173,7 +198,9 @@ identifier = do
 
     let indexed = do
             _ <- Text.Parser.Char.char '@'
-            Text.Parser.Token.natural
+            n <- Text.Megaparsec.Char.Lexer.decimal
+            whitespace
+            return n
 
     n <- indexed <|> pure 0
     return (V x n)
@@ -194,19 +221,19 @@ hexNumber = choice [ hexDigit, hexUpper, hexLower ]
   where
     hexDigit = do
         c <- Text.Parser.Char.satisfy predicate
-        return (Data.Char.ord c - Data.Char.ord '0')
+        return (Char.ord c - Char.ord '0')
       where
         predicate c = '0' <= c && c <= '9'
 
     hexUpper = do
         c <- Text.Parser.Char.satisfy predicate
-        return (10 + Data.Char.ord c - Data.Char.ord 'A')
+        return (10 + Char.ord c - Char.ord 'A')
       where
         predicate c = 'A' <= c && c <= 'F'
 
     hexLower = do
         c <- Text.Parser.Char.satisfy predicate
-        return (10 + Data.Char.ord c - Data.Char.ord 'a')
+        return (10 + Char.ord c - Char.ord 'a')
       where
         predicate c = 'a' <= c && c <= 'f'
 
@@ -245,7 +272,7 @@ blockCommentChunk =
         predicate c =
                 '\x20' <= c && c <= '\x10FFFF' && c /= '-' && c /= '{'
             ||  c == '\n'
-            || c == '\t'
+            ||  c == '\t'
 
     character = void (Text.Parser.Char.satisfy predicate)
       where
@@ -262,17 +289,17 @@ blockCommentContinue = endOfComment <|> continue
         blockCommentChunk
         blockCommentContinue
 
-simpleLabel :: Parser Text
-simpleLabel = try (do
+simpleLabel :: Bool -> Parser Text
+simpleLabel allowReserved = try (do
     c    <- Text.Parser.Char.satisfy headCharacter
     rest <- Dhall.Parser.Combinators.takeWhile tailCharacter
     let text = Data.Text.cons c rest
-    Control.Monad.guard (not (Data.HashSet.member text reservedIdentifiers))
+    Control.Monad.guard (allowReserved || not (Data.HashSet.member text reservedIdentifiers))
     return text )
   where
     headCharacter c = alpha c || c == '_'
 
-    tailCharacter c = alpha c || digit c || c == '_' || c == '-' || c == '/'
+    tailCharacter c = alphaNum c || c == '_' || c == '-' || c == '/'
 
 backtickLabel :: Parser Text
 backtickLabel = do
@@ -281,7 +308,9 @@ backtickLabel = do
     _ <- Text.Parser.Char.char '`'
     return t
   where
-    predicate c = alpha c || digit c || elem c ("$-/_:." :: String)
+    predicate c =
+            '\x20' <= c && c <= '\x5F'
+        ||  '\x61' <= c && c <= '\x7E'
 
 labels :: Parser (Set Text)
 labels = do
@@ -293,32 +322,56 @@ labels = do
     emptyLabels = pure Dhall.Set.empty
 
     nonEmptyLabels = do
-        x  <- label
-        xs <- many (do _ <- _comma; label)
+        x  <- anyLabel
+        xs <- many (do _ <- _comma; anyLabel)
         noDuplicates (x : xs)
 
 
+labelOnly :: Parser Text
+labelOnly = backtickLabel <|> simpleLabel False <?> "label"
+
 label :: Parser Text
 label = (do
-    t <- backtickLabel <|> simpleLabel
+    t <- backtickLabel <|> simpleLabel False
     whitespace
     return t ) <?> "label"
 
+anyLabel :: Parser Text
+anyLabel = (do
+    t <- backtickLabel <|> simpleLabel True
+    whitespace
+    return t ) <?> "any label"
+
 bashEnvironmentVariable :: Parser Text
 bashEnvironmentVariable = satisfy predicate0 <> star (satisfy predicate1)
   where
     predicate0 c = alpha c || c == '_'
 
-    predicate1 c = alpha c || digit c || c == '_'
+    predicate1 c = alphaNum c || c == '_'
 
 posixEnvironmentVariable :: Parser Text
 posixEnvironmentVariable = plus posixEnvironmentVariableCharacter
 
 posixEnvironmentVariableCharacter :: Parser Text
 posixEnvironmentVariableCharacter =
-    ("\\" <> satisfy predicate0) <|> satisfy predicate1
+    escapeCharacter <|> satisfy predicate1
   where
-    predicate0 c = c `elem` ("\"\\abfnrtv" :: String)
+    escapeCharacter = do
+        _ <- Text.Parser.Char.char '\\'
+
+        c <- Text.Parser.Char.satisfy (`elem` ("\"\\abfnrtv" :: String))
+
+        case c of
+            '"'  -> return "\""
+            '\\' -> return "\\"
+            'a'  -> return "\a"
+            'b'  -> return "\b"
+            'f'  -> return "\f"
+            'n'  -> return "\n"
+            'r'  -> return "\r"
+            't'  -> return "\t"
+            'v'  -> return "\v"
+            _    -> empty
 
     predicate1 c =
             ('\x20' <= c && c <= '\x21')
@@ -330,25 +383,42 @@ quotedPathCharacter :: Char -> Bool
 quotedPathCharacter c =
         ('\x20' <= c && c <= '\x21')
     ||  ('\x23' <= c && c <= '\x2E')
-    ||  ('\x30' <= c && c <= '\x7E')
+    ||  ('\x30' <= c && c <= '\x10FFFF')
 
-pathComponent :: Parser Text
-pathComponent = do
-    _      <- "/" :: Parser Text
+data ComponentType = URLComponent | FileComponent
 
-    let pathData = Text.Megaparsec.takeWhile1P Nothing Dhall.Core.pathCharacter
+pathComponent :: ComponentType -> Parser Text
+pathComponent componentType = do
+    _ <- "/" :: Parser Text
+
+    let pathData =
+            case componentType of
+                FileComponent -> do
+                    Text.Megaparsec.takeWhile1P Nothing Dhall.Core.pathCharacter
+                URLComponent -> do
+                    star pchar
 
     let quotedPathData = do
             _    <- Text.Parser.Char.char '"'
             text <- Text.Megaparsec.takeWhile1P Nothing quotedPathCharacter
             _    <- Text.Parser.Char.char '"'
-            return text
 
-    pathData <|> quotedPathData
+            case componentType of
+              FileComponent -> do
+                return text
+              URLComponent -> do
+                return (URI.Encode.encodeText text)
 
-file_ :: Parser File
-file_ = do
-    path <- Data.List.NonEmpty.some1 pathComponent
+    quotedPathData <|> pathData
+
+file_ :: ComponentType -> Parser File
+file_ componentType = do
+    let emptyPath =
+            case componentType of
+                URLComponent  -> pure (pure "")
+                FileComponent -> empty
+
+    path <- Data.List.NonEmpty.some1 (pathComponent componentType) <|> emptyPath
 
     let directory = Directory (reverse (Data.List.NonEmpty.init path))
     let file      = Data.List.NonEmpty.last path
@@ -365,9 +435,8 @@ httpRaw :: Parser URL
 httpRaw = do
     scheme    <- scheme_
     authority <- authority_
-    path      <- file_
+    path      <- file_ URLComponent
     query     <- optional (("?" :: Parser Text) *> query_)
-    fragment  <- optional (("#" :: Parser Text) *> fragment_)
 
     let headers = Nothing
 
@@ -382,7 +451,7 @@ userinfo = star (satisfy predicate <|> pctEncoded)
     predicate c = unreserved c || subDelims c || c == ':'
 
 host :: Parser Text
-host = choice [ ipLiteral, ipV4Address, regName ]
+host = choice [ ipLiteral, try ipV4Address, domain ]
 
 port :: Parser Text
 port = star (satisfy digit)
@@ -416,28 +485,32 @@ ipV6Address =
     alternative2 = option h16 <> "::" <> count 4 (h16 <> ":") <> ls32
 
     alternative3 =
-            option (range 0 1 (h16 <> ":") <> h16)
+            option (h16 <> range 0 1 (try (":" <> h16)))
         <>  "::"
         <>  count 3 (h16 <> ":")
         <>  ls32
 
     alternative4 =
-            option (range 0 2 (h16 <> ":") <> h16)
+            option (h16 <> range 0 2 (try (":" <> h16)))
         <>  "::"
         <>  count 2 (h16 <> ":")
         <>  ls32
 
     alternative5 =
-        option (range 0 3 (h16 <> ":") <> h16) <> "::" <> h16 <> ":" <> ls32
+            option (h16 <> range 0 3 (try (":" <> h16)))
+        <>  "::"
+        <>  h16
+        <>  ":"
+        <>  ls32
 
     alternative6 =
-        option (range 0 4 (h16 <> ":") <> h16) <> "::" <> ls32
+        option (h16 <> range 0 4 (try (":" <> h16))) <> "::" <> ls32
 
     alternative7 =
-        option (range 0 5 (h16 <> ":") <> h16) <> "::" <> h16
+        option (h16 <> range 0 5 (try (":" <> h16))) <> "::" <> h16
 
     alternative8 =
-        option (range 0 6 (h16 <> ":") <> h16) <> "::"
+        option (h16 <> range 0 6 (try (":" <> h16))) <> "::"
 
 h16 :: Parser Text
 h16 = range 1 3 (satisfy hexdig)
@@ -474,10 +547,13 @@ decOctet =
       where
         predicate c = '\x30' <= c && c <= '\x35'
 
-regName :: Parser Text
-regName = star (satisfy predicate <|> pctEncoded)
+domain :: Parser Text
+domain = domainLabel <> star ("." <> domainLabel ) <> option "."
+
+domainLabel :: Parser Text
+domainLabel = plus alphaNum_ <> star (plus "-" <> plus alphaNum_)
   where
-    predicate c = unreserved c || subDelims c
+    alphaNum_ = satisfy alphaNum
 
 pchar :: Parser Text
 pchar = satisfy predicate <|> pctEncoded
@@ -489,11 +565,6 @@ query_ = star (pchar <|> satisfy predicate)
   where
     predicate c = c == '/' || c == '?'
 
-fragment_ :: Parser Text
-fragment_ = star (pchar <|> satisfy predicate)
-  where
-    predicate c = c == '/' || c == '?'
-
 pctEncoded :: Parser Text
 pctEncoded = "%" <> count 2 (satisfy hexdig)
 
@@ -502,14 +573,20 @@ subDelims c = c `elem` ("!$&'()*+,;=" :: String)
 
 unreserved :: Char -> Bool
 unreserved c =
-    alpha c || digit c || c == '-' || c == '.' || c == '_' || c == '~'
+    alphaNum c || c == '-' || c == '.' || c == '_' || c == '~'
 
 reserved :: Data.Text.Text -> Parser ()
 reserved x = do _ <- Text.Parser.Char.text x; whitespace
 
+reservedCharOnly :: Char -> Parser ()
+reservedCharOnly c = do _ <- Text.Parser.Char.char c; return ()
+
 reservedChar :: Char -> Parser ()
 reservedChar c = do _ <- Text.Parser.Char.char c; whitespace
 
+keywordOnly :: Data.Text.Text -> Parser ()
+keywordOnly x = try (do _ <- Text.Parser.Char.text x; return ())
+
 keyword :: Data.Text.Text -> Parser ()
 keyword x = try (do _ <- Text.Parser.Char.text x; nonemptyWhitespace)
 
@@ -522,6 +599,9 @@ _then = keyword "then"
 _else :: Parser ()
 _else = keyword "else"
 
+_letOnly :: Parser ()
+_letOnly = keywordOnly "let"
+
 _let :: Parser ()
 _let = keyword "let"
 
@@ -537,8 +617,11 @@ _using = keyword "using"
 _merge :: Parser ()
 _merge = keyword "merge"
 
-_constructors :: Parser ()
-_constructors = keyword "constructors"
+_toMap :: Parser ()
+_toMap = keyword "toMap"
+
+_assert :: Parser ()
+_assert = keyword "assert"
 
 _Some :: Parser ()
 _Some = keyword "Some"
@@ -567,6 +650,9 @@ _NaturalToInteger = reserved "Natural/toInteger"
 _NaturalShow :: Parser ()
 _NaturalShow = reserved "Natural/show"
 
+_NaturalSubtract :: Parser ()
+_NaturalSubtract = reserved "Natural/subtract"
+
 _IntegerShow :: Parser ()
 _IntegerShow = reserved "Integer/show"
 
@@ -621,6 +707,9 @@ _Double = reserved "Double"
 _Text :: Parser ()
 _Text = reserved "Text"
 
+_TextShow :: Parser ()
+_TextShow = reserved "Text/show"
+
 _List :: Parser ()
 _List = reserved "List"
 
@@ -642,6 +731,12 @@ _Kind = reserved "Kind"
 _Sort :: Parser ()
 _Sort = reserved "Sort"
 
+_Location :: Parser ()
+_Location = reserved "Location"
+
+_equalOnly :: Parser ()
+_equalOnly = reservedCharOnly '='
+
 _equal :: Parser ()
 _equal = reservedChar '='
 
@@ -702,12 +797,20 @@ _openParens = reservedChar '('
 _closeParens :: Parser ()
 _closeParens = reservedChar ')'
 
+_colonOnly :: Parser ()
+_colonOnly = reservedCharOnly ':'
+
 _colon :: Parser ()
 _colon = reservedChar ':'
 
 _at :: Parser ()
 _at = reservedChar '@'
 
+_equivalent :: Parser ()
+_equivalent = do
+    void (Text.Parser.Char.char '≡' <?> "\"≡\"") <|> void (Text.Parser.Char.text "===")
+    whitespace
+
 _missing :: Parser ()
 _missing = reserved "missing"
 
