@@ -1,5 +1,7 @@
 {-# LANGUAGE CPP               #-}
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE RecordWildCards   #-}
+
 {-# OPTIONS_GHC -Wall #-}
 
 {-| This module provides internal pretty-printing utilities which are used by
@@ -10,6 +12,7 @@ module Dhall.Pretty.Internal (
       Ann(..)
     , annToAnsiStyle
     , prettyExpr
+    , prettySrcExpr
 
     , CharacterSet(..)
     , prettyCharacterSet
@@ -20,9 +23,11 @@ module Dhall.Pretty.Internal (
 
     , prettyConst
     , prettyLabel
+    , prettyAnyLabel
     , prettyLabels
     , prettyNatural
     , prettyNumber
+    , prettyInt
     , prettyDouble
     , prettyToStrictText
     , prettyToString
@@ -65,6 +70,7 @@ import Data.Text (Text)
 import Data.Text.Prettyprint.Doc (Doc, Pretty, space)
 import Dhall.Map (Map)
 import Dhall.Set (Set)
+import Dhall.Src (Src(..))
 import Numeric.Natural (Natural)
 import Prelude hiding (succ)
 import qualified Data.Text.Prettyprint.Doc.Render.Terminal as Terminal
@@ -106,7 +112,10 @@ data CharacterSet = ASCII | Unicode
 
 -- | Pretty print an expression
 prettyExpr :: Pretty a => Expr s a -> Doc Ann
-prettyExpr = prettyCharacterSet Unicode
+prettyExpr = prettySrcExpr . denote
+
+prettySrcExpr :: Pretty a => Expr Src a -> Doc Ann
+prettySrcExpr = prettyCharacterSet Unicode
 
 {-| Internal utility for pretty-printing, used when generating element lists
     to supply to `enclose` or `enclose'`.  This utility indicates that the
@@ -116,6 +125,79 @@ prettyExpr = prettyCharacterSet Unicode
 duplicate :: a -> (a, a)
 duplicate x = (x, x)
 
+isWhitespace :: Char -> Bool
+isWhitespace c =
+    case c of
+        ' '  -> True
+        '\n' -> True
+        '\t' -> True
+        '\r' -> True
+        _    -> False
+
+{-| Used to render inline `Src` spans preserved by the syntax tree
+
+    >>> let unusedSourcePos = Text.Megaparsec.SourcePos "" (Text.Megaparsec.mkPos 1) (Text.Megaparsec.mkPos 1)
+    >>> let nonEmptySrc = Src unusedSourcePos unusedSourcePos "-- Documentation for x\n"
+    >>> "let" <> renderSrc (Just nonEmptySrc) " " <> "x = 1 in x"
+    let -- Documentation for x
+        x = 1 in x
+    >>> let emptySrc = Src unusedSourcePos unusedSourcePos "      "
+    >>> "let" <> renderSrc (Just emptySrc) " " <> "x = 1 in x"
+    let x = 1 in x
+    >>> "let" <> renderSrc Nothing " " <> "x = 1 in x"
+    let x = 1 in x
+-}
+renderSrc
+    :: Maybe Src
+    -- ^ Source span to render (if present)
+    -> Doc Ann
+    -- ^ Used as the prefix (when the source span contains a comment) and as a
+    --   fallback (when the source span is absent or comment-free)
+    -> Doc Ann
+renderSrc (Just (Src {..})) prefix
+    | not (Text.all isWhitespace srcText) =
+        prefix <> Pretty.align (Pretty.concatWith f newLines <> suffix)
+  where
+    horizontalSpace c = c == ' ' || c == '\t'
+
+    strippedText = Text.dropAround horizontalSpace srcText
+
+    suffix =
+        if Text.null strippedText
+        then mempty
+        else if Text.last strippedText == '\n' then mempty else " "
+
+    oldLines = Text.splitOn "\n" strippedText
+
+    spacePrefix = Text.takeWhile horizontalSpace
+
+    commonPrefix a b = case Text.commonPrefixes a b of
+        Nothing        -> ""
+        Just (c, _, _) -> c
+
+    blank = Text.all horizontalSpace
+
+    newLines =
+        case oldLines of
+            [] ->
+               []
+            l0 : [] ->
+                Pretty.pretty l0 : []
+            l0 : l1 : ls ->
+                let sharedPrefix =
+                        foldl' commonPrefix (spacePrefix l1) (map spacePrefix (filter (not . blank) ls))
+
+                    perLine l =
+                        case Text.stripPrefix sharedPrefix l of
+                            Nothing -> Pretty.pretty l
+                            Just l' -> Pretty.pretty l'
+
+                in  Pretty.pretty l0 : map perLine (l1 : ls)
+
+    f x y = x <> Pretty.hardline <> y
+renderSrc _ prefix =
+    prefix
+
 -- Annotation helpers
 keyword, syntax, label, literal, builtin, operator :: Doc Ann -> Doc Ann
 keyword  = Pretty.annotate Keyword
@@ -195,12 +277,34 @@ braces docs =
 
 -- | Pretty-print anonymous functions and function types
 arrows :: CharacterSet -> [(Doc Ann, Doc Ann)] -> Doc Ann
-arrows characterSet =
+arrows ASCII =
+    enclose'
+        ""
+        "    "
+        (" " <> rarrow ASCII <> " ")
+        (rarrow ASCII <> "  ")
+arrows Unicode =
     enclose'
         ""
         "  "
-        (" " <> rarrow characterSet <> " ")
-        (rarrow characterSet <> space)
+        (" " <> rarrow Unicode <> " ")
+        (rarrow Unicode <> " ")
+
+combine :: CharacterSet -> Text
+combine ASCII   = "/\\"
+combine Unicode = "∧"
+
+combineTypes :: CharacterSet -> Text
+combineTypes ASCII   = "//\\\\"
+combineTypes Unicode = "⩓"
+
+prefer :: CharacterSet -> Text
+prefer ASCII   = "//"
+prefer Unicode = "⫽"
+
+equivalent :: CharacterSet -> Text
+equivalent ASCII   = "==="
+equivalent Unicode = "≡"
 
 {-| Format an expression that holds a variable number of elements, such as a
     list, record, or union
@@ -282,26 +386,35 @@ headCharacter c = alpha c || c == '_'
 tailCharacter :: Char -> Bool
 tailCharacter c = alpha c || digit c || c == '_' || c == '-' || c == '/'
 
-prettyLabel :: Text -> Doc Ann
-prettyLabel a = label doc
+prettyLabelShared :: Bool -> Text -> Doc Ann
+prettyLabelShared allowReserved a = label doc
     where
         doc =
             case Text.uncons a of
                 Just (h, t)
-                    | headCharacter h && Text.all tailCharacter t && not (Data.HashSet.member a reservedIdentifiers)
+                    | headCharacter h && Text.all tailCharacter t && (allowReserved || not (Data.HashSet.member a reservedIdentifiers))
                         -> Pretty.pretty a
                 _       -> backtick <> Pretty.pretty a <> backtick
 
+prettyLabel :: Text -> Doc Ann
+prettyLabel = prettyLabelShared False
+
+prettyAnyLabel :: Text -> Doc Ann
+prettyAnyLabel = prettyLabelShared True
+
 prettyLabels :: Set Text -> Doc Ann
 prettyLabels a
     | Data.Set.null (Dhall.Set.toSet a) =
         lbrace <> rbrace
     | otherwise =
-        braces (map (duplicate . prettyLabel) (Dhall.Set.toList a))
+        braces (map (duplicate . prettyAnyLabel) (Dhall.Set.toList a))
 
 prettyNumber :: Integer -> Doc Ann
 prettyNumber = literal . Pretty.pretty
 
+prettyInt :: Int -> Doc Ann
+prettyInt = literal . Pretty.pretty
+
 prettyNatural :: Natural -> Doc Ann
 prettyNatural = literal . Pretty.pretty
 
@@ -315,10 +428,24 @@ prettyConst Sort = builtin "Sort"
 
 prettyVar :: Var -> Doc Ann
 prettyVar (V x 0) = label (Pretty.unAnnotate (prettyLabel x))
-prettyVar (V x n) = label (Pretty.unAnnotate (prettyLabel x <> "@" <> prettyNumber n))
+prettyVar (V x n) = label (Pretty.unAnnotate (prettyLabel x <> "@" <> prettyInt n))
 
-prettyCharacterSet :: Pretty a => CharacterSet -> Expr s a -> Doc Ann
-prettyCharacterSet characterSet = prettyExpression
+{-| Pretty-print an 'Expr' using the given 'CharacterSet'.
+
+'prettyCharacterSet' largely ignores 'Note's. 'Note's do however matter for
+the layout of let-blocks:
+
+>>> let inner = Let (Binding Nothing "x" Nothing Nothing Nothing (NaturalLit 1)) (Var (V "x" 0)) :: Expr Src ()
+>>> prettyCharacterSet ASCII (Let (Binding Nothing "y" Nothing Nothing Nothing (NaturalLit 2)) inner)
+let y = 2 let x = 1 in x
+>>> prettyCharacterSet ASCII (Let (Binding Nothing "y" Nothing Nothing Nothing (NaturalLit 2)) (Note (Src unusedSourcePos unusedSourcePos "") inner))
+let y = 2 in let x = 1 in x
+
+This means the structure of parsed let-blocks is preserved.
+-}
+prettyCharacterSet :: Pretty a => CharacterSet -> Expr Src a -> Doc Ann
+prettyCharacterSet characterSet expression =
+    Pretty.group (prettyExpression expression)
   where
     prettyExpression a0@(Lam _ _ _) =
         arrows characterSet (fmap duplicate (docs a0))
@@ -386,45 +513,42 @@ prettyCharacterSet characterSet = prettyExpression
                 ]
         docsShort (Note  _    c) = docsShort c
         docsShort             c  = [ prettyExpression c ]
-    prettyExpression (Let as b) =
+    prettyExpression (Let a0 b0) =
         enclose' "" "" space Pretty.hardline
             (fmap duplicate (fmap docA (toList as)) ++ [ docB ])
       where
-        docA (Binding c Nothing e) =
+        MultiLet as b = multiLet a0 b0
+
+        docA (Binding src0 c src1 Nothing src2 e) =
             Pretty.group (Pretty.flatAlt long short)
           where
             long =  keyword "let" <> space
                 <>  Pretty.align
-                    (   prettyLabel c
-                    <>  space <> equals
-                    <>  Pretty.hardline
-                    <>  "  "
-                    <>  prettyExpression e
+                    (   renderSrc src0 mempty
+                    <>  prettyLabel c <> renderSrc src1 space
+                    <>  equals <> renderSrc src2 Pretty.hardline
+                    <>  "  " <> prettyExpression e
                     )
 
-            short = keyword "let" <> space
-                <>  prettyLabel c
-                <>  (space <> equals <> space)
+            short = keyword "let" <> renderSrc src0 space
+                <>  prettyLabel c <> renderSrc src1 space
+                <>  equals <> renderSrc src2 space
                 <>  prettyExpression e
-        docA (Binding c (Just d) e) =
+        docA (Binding src0 c src1 (Just (src3, d)) src2 e) =
             Pretty.group (Pretty.flatAlt long short)
           where
             long = keyword "let" <> space
                 <>  Pretty.align
-                    (   prettyLabel c
-                    <>  Pretty.hardline
-                    <>  colon <> space
-                    <>  prettyExpression d
-                    <>  Pretty.hardline
-                    <>  equals <> space
+                    (   renderSrc src0 mempty
+                    <>  prettyLabel c <> renderSrc src1 Pretty.hardline
+                    <>  colon <> renderSrc src3 space <> prettyExpression d <> Pretty.hardline <> equals <> renderSrc src2 space
                     <>  prettyExpression e
                     )
 
-            short = keyword "let" <> space
-                <>  prettyLabel c
-                <>  space <> colon <> space
-                <>  prettyExpression d
-                <>  space <> equals <> space
+            short = keyword "let" <> renderSrc src0 space
+                <>  prettyLabel c <> renderSrc src1 space
+                <>  colon <> renderSrc src3 space
+                <>  prettyExpression d <> space <> equals <> renderSrc src2 space
                 <>  prettyExpression e
 
         docB =
@@ -455,12 +579,22 @@ prettyCharacterSet characterSet = prettyExpression
                 <>  rparen
         docs (Note _   c) = docs c
         docs           c  = [ prettyExpression c ]
+    prettyExpression (Assert a) =
+        Pretty.group (Pretty.flatAlt long short)
+      where
+        short = keyword "assert" <> " " <> colon <> " " <> prettyExpression a
+
+        long =
+            Pretty.align
+            (  "  " <> keyword "assert"
+            <> Pretty.hardline <> colon <> " " <> prettyExpression a
+            )
     prettyExpression (Note _ a) =
         prettyExpression a
     prettyExpression a0 =
         prettyAnnotatedExpression a0
 
-    prettyAnnotatedExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyAnnotatedExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyAnnotatedExpression (Merge a b (Just c)) =
         Pretty.group (Pretty.flatAlt long short)
       where
@@ -468,9 +602,9 @@ prettyCharacterSet characterSet = prettyExpression
             Pretty.align
                 (   keyword "merge"
                 <>  Pretty.hardline
-                <>  prettyImportExpression a
+                <>  Pretty.indent 2 (prettyImportExpression a)
                 <>  Pretty.hardline
-                <>  prettyImportExpression b
+                <>  Pretty.indent 2 (prettyImportExpression b)
                 <>  Pretty.hardline
                 <>  colon <> space
                 <>  prettyApplicationExpression c
@@ -489,15 +623,44 @@ prettyCharacterSet characterSet = prettyExpression
             Pretty.align
                 (   keyword "merge"
                 <>  Pretty.hardline
-                <>  prettyImportExpression a
+                <>  Pretty.indent 2 (prettyImportExpression a)
                 <>  Pretty.hardline
-                <>  prettyImportExpression b
+                <>  Pretty.indent 2 (prettyImportExpression b)
                 )
 
         short = keyword "merge" <> space
             <>  prettyImportExpression a
             <>  " "
             <>  prettyImportExpression b
+    prettyAnnotatedExpression (ToMap a (Just b)) =
+        Pretty.group (Pretty.flatAlt long short)
+      where
+        long =
+            Pretty.align
+                (   keyword "toMap"
+                <>  Pretty.hardline
+                <>  Pretty.indent 2 (prettyImportExpression a)
+                <>  Pretty.hardline
+                <>  colon <> space
+                <>  prettyApplicationExpression b
+                )
+
+        short = keyword "toMap" <> space
+            <>  prettyImportExpression a
+            <>  space <> colon <> space
+            <>  prettyApplicationExpression b
+    prettyAnnotatedExpression (ToMap a Nothing) =
+        Pretty.group (Pretty.flatAlt long short)
+      where
+        long =
+            Pretty.align
+                (   keyword "toMap"
+                <>  Pretty.hardline
+                <>  Pretty.indent 2 (prettyImportExpression a)
+                )
+
+        short = keyword "toMap" <> space
+            <>  prettyImportExpression a
     prettyAnnotatedExpression a0@(Annot _ _) =
         enclose'
             ""
@@ -512,17 +675,13 @@ prettyCharacterSet characterSet = prettyExpression
     prettyAnnotatedExpression (ListLit (Just a) b) =
             list (map prettyExpression (Data.Foldable.toList b))
         <>  " : "
-        <>  prettyApplicationExpression (App List a)
-    prettyAnnotatedExpression (OptionalLit a b) =
-            list (map prettyExpression (Data.Foldable.toList b))
-        <>  " : "
-        <>  prettyApplicationExpression (App Optional a)
+        <>  prettyApplicationExpression a
     prettyAnnotatedExpression (Note _ a) =
         prettyAnnotatedExpression a
     prettyAnnotatedExpression a0 =
         prettyOperatorExpression a0
 
-    prettyOperatorExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyOperatorExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyOperatorExpression = prettyImportAltExpression
 
     prettyOperator :: Text -> [Doc Ann] -> Doc Ann
@@ -538,7 +697,7 @@ prettyCharacterSet characterSet = prettyExpression
 
         spacer = if Text.length op == 1 then " "  else "  "
 
-    prettyImportAltExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyImportAltExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyImportAltExpression a0@(ImportAlt _ _) =
         prettyOperator "?" (docs a0)
       where
@@ -550,7 +709,7 @@ prettyCharacterSet characterSet = prettyExpression
     prettyImportAltExpression a0 =
         prettyOrExpression a0
 
-    prettyOrExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyOrExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyOrExpression a0@(BoolOr _ _) =
         prettyOperator "||" (docs a0)
       where
@@ -562,7 +721,7 @@ prettyCharacterSet characterSet = prettyExpression
     prettyOrExpression a0 =
         prettyPlusExpression a0
 
-    prettyPlusExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyPlusExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyPlusExpression a0@(NaturalPlus _ _) =
         prettyOperator "+" (docs a0)
       where
@@ -574,7 +733,7 @@ prettyCharacterSet characterSet = prettyExpression
     prettyPlusExpression a0 =
         prettyTextAppendExpression a0
 
-    prettyTextAppendExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyTextAppendExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyTextAppendExpression a0@(TextAppend _ _) =
         prettyOperator "++" (docs a0)
       where
@@ -586,7 +745,7 @@ prettyCharacterSet characterSet = prettyExpression
     prettyTextAppendExpression a0 =
         prettyListAppendExpression a0
 
-    prettyListAppendExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyListAppendExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyListAppendExpression a0@(ListAppend _ _) =
         prettyOperator "#" (docs a0)
       where
@@ -598,7 +757,7 @@ prettyCharacterSet characterSet = prettyExpression
     prettyListAppendExpression a0 =
         prettyAndExpression a0
 
-    prettyAndExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyAndExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyAndExpression a0@(BoolAnd _ _) =
         prettyOperator "&&" (docs a0)
       where
@@ -610,9 +769,9 @@ prettyCharacterSet characterSet = prettyExpression
     prettyAndExpression a0 =
        prettyCombineExpression a0
 
-    prettyCombineExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyCombineExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyCombineExpression a0@(Combine _ _) =
-        prettyOperator "∧" (docs a0)
+        prettyOperator (combine characterSet) (docs a0)
       where
         docs (Combine a b) = prettyPreferExpression b : docs a
         docs (Note    _ b) = docs b
@@ -622,9 +781,9 @@ prettyCharacterSet characterSet = prettyExpression
     prettyCombineExpression a0 =
         prettyPreferExpression a0
 
-    prettyPreferExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyPreferExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyPreferExpression a0@(Prefer _ _) =
-        prettyOperator "⫽" (docs a0)
+        prettyOperator (prefer characterSet) (docs a0)
       where
         docs (Prefer a b) = prettyCombineTypesExpression b : docs a
         docs (Note   _ b) = docs b
@@ -634,9 +793,9 @@ prettyCharacterSet characterSet = prettyExpression
     prettyPreferExpression a0 =
         prettyCombineTypesExpression a0
 
-    prettyCombineTypesExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyCombineTypesExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyCombineTypesExpression a0@(CombineTypes _ _) =
-        prettyOperator "⩓" (docs a0)
+        prettyOperator (combineTypes characterSet) (docs a0)
       where
         docs (CombineTypes a b) = prettyTimesExpression b : docs a
         docs (Note         _ b) = docs b
@@ -646,7 +805,7 @@ prettyCharacterSet characterSet = prettyExpression
     prettyCombineTypesExpression a0 =
         prettyTimesExpression a0
 
-    prettyTimesExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyTimesExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyTimesExpression a0@(NaturalTimes _ _) =
         prettyOperator "*" (docs a0)
       where
@@ -658,7 +817,7 @@ prettyCharacterSet characterSet = prettyExpression
     prettyTimesExpression a0 =
         prettyEqualExpression a0
 
-    prettyEqualExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyEqualExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyEqualExpression a0@(BoolEQ _ _) =
         prettyOperator "==" (docs a0)
       where
@@ -670,35 +829,45 @@ prettyCharacterSet characterSet = prettyExpression
     prettyEqualExpression a0 =
         prettyNotEqualExpression a0
 
-    prettyNotEqualExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyNotEqualExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyNotEqualExpression a0@(BoolNE _ _) =
         prettyOperator "!=" (docs a0)
       where
-        docs (BoolNE a b) = prettyApplicationExpression b : docs a
+        docs (BoolNE a b) = prettyEquivalentExpression b : docs a
         docs (Note   _ b) = docs b
-        docs           b  = [ prettyApplicationExpression b ]
+        docs           b  = [ prettyEquivalentExpression b ]
     prettyNotEqualExpression (Note _ a) =
         prettyNotEqualExpression a
     prettyNotEqualExpression a0 =
+        prettyEquivalentExpression a0
+
+    prettyEquivalentExpression :: Pretty a => Expr Src a -> Doc Ann
+    prettyEquivalentExpression a0@(Equivalent _ _) =
+        prettyOperator (equivalent characterSet) (docs a0)
+      where
+        docs (Equivalent a b) = prettyApplicationExpression b : docs a
+        docs (Note       _ b) = docs b
+        docs               b  = [ prettyApplicationExpression b ]
+    prettyEquivalentExpression (Note _ a) =
+        prettyEquivalentExpression a
+    prettyEquivalentExpression a0 =
         prettyApplicationExpression a0
 
-    prettyApplicationExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyApplicationExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyApplicationExpression a0 = case a0 of
-        App _ _        -> result
-        Constructors _ -> result
-        Some _         -> result
-        Note _ b       -> prettyApplicationExpression b
-        _              -> prettyImportExpression a0
+        App _ _  -> result
+        Some _   -> result
+        Note _ b -> prettyApplicationExpression b
+        _        -> prettyImportExpression a0
       where
-        result = enclose' "" "" " " "" (fmap duplicate (reverse (docs a0)))
+        result = enclose' "" "" " " "" (reverse (docs a0))
 
-        docs (App        a b) = prettyImportExpression b : docs a
-        docs (Constructors b) = [ prettyImportExpression b , keyword "constructors" ]
-        docs (Some         a) = [ prettyImportExpression a , builtin "Some"         ]
-        docs (Note       _ b) = docs b
-        docs               b  = [ prettyImportExpression b ]
+        docs (App  a b) = ( prettyImportExpression b, Pretty.indent 2 (prettyImportExpression b) ) : docs a
+        docs (Some   a) = map duplicate [ prettyImportExpression a , builtin "Some" ]
+        docs (Note _ b) = docs b
+        docs         b  = map duplicate [ prettyImportExpression b ]
 
-    prettyImportExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyImportExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyImportExpression (Embed a) =
         Pretty.pretty a
     prettyImportExpression (Note _ a) =
@@ -706,17 +875,23 @@ prettyCharacterSet characterSet = prettyExpression
     prettyImportExpression a0 =
         prettySelectorExpression a0
 
-    prettySelectorExpression :: Pretty a => Expr s a -> Doc Ann
+    prettySelectorExpression :: Pretty a => Expr Src a -> Doc Ann
     prettySelectorExpression (Field a b) =
-        prettySelectorExpression a <> dot <> prettyLabel b
-    prettySelectorExpression (Project a b) =
+        prettySelectorExpression a <> dot <> prettyAnyLabel b
+    prettySelectorExpression (Project a (Left b)) =
         prettySelectorExpression a <> dot <> prettyLabels b
+    prettySelectorExpression (Project a (Right b)) =
+            prettySelectorExpression a
+        <>  dot
+        <>  lparen
+        <>  prettyExpression b
+        <>  rparen
     prettySelectorExpression (Note _ b) =
         prettySelectorExpression b
     prettySelectorExpression a0 =
         prettyPrimitiveExpression a0
 
-    prettyPrimitiveExpression :: Pretty a => Expr s a -> Doc Ann
+    prettyPrimitiveExpression :: Pretty a => Expr Src a -> Doc Ann
     prettyPrimitiveExpression (Var a) =
         prettyVar a
     prettyPrimitiveExpression (Const k) =
@@ -739,6 +914,8 @@ prettyCharacterSet characterSet = prettyExpression
         builtin "Natural/toInteger"
     prettyPrimitiveExpression NaturalShow =
         builtin "Natural/show"
+    prettyPrimitiveExpression NaturalSubtract =
+        builtin "Natural/subtract"
     prettyPrimitiveExpression Integer =
         builtin "Integer"
     prettyPrimitiveExpression IntegerShow =
@@ -751,6 +928,8 @@ prettyCharacterSet characterSet = prettyExpression
         builtin "Double/show"
     prettyPrimitiveExpression Text =
         builtin "Text"
+    prettyPrimitiveExpression TextShow =
+        builtin "Text/show"
     prettyPrimitiveExpression List =
         builtin "List"
     prettyPrimitiveExpression ListBuild =
@@ -794,8 +973,6 @@ prettyCharacterSet characterSet = prettyExpression
         prettyRecordLit a
     prettyPrimitiveExpression (Union a) =
         prettyUnion a
-    prettyPrimitiveExpression (UnionLit a b c) =
-        prettyUnionLit a b c
     prettyPrimitiveExpression (ListLit Nothing b) =
         list (map prettyExpression (Data.Foldable.toList b))
     prettyPrimitiveExpression (Note _ b) =
@@ -809,42 +986,42 @@ prettyCharacterSet characterSet = prettyExpression
 
         short = lparen <> prettyExpression a <> rparen
 
-    prettyKeyValue :: Pretty a => Doc Ann -> (Text, Expr s a) -> (Doc Ann, Doc Ann)
+    prettyKeyValue :: Pretty a => Doc Ann -> (Text, Expr Src a) -> (Doc Ann, Doc Ann)
     prettyKeyValue separator (key, val) =
-        (   prettyLabel key <> " " <> separator <> " " <> prettyExpression val
-        ,       prettyLabel key
+        duplicate (Pretty.group (Pretty.flatAlt long short))
+      where
+        short = prettyAnyLabel key
             <>  " "
             <>  separator
-            <>  long
-        )
-      where
-        long = Pretty.hardline <> "    " <> prettyExpression val
+            <>  " "
+            <>  prettyExpression val
 
-    prettyRecord :: Pretty a => Map Text (Expr s a) -> Doc Ann
+        long =  prettyAnyLabel key
+            <>  " "
+            <>  separator
+            <>  Pretty.hardline
+            <>  "    "
+            <>  prettyExpression val
+
+    prettyRecord :: Pretty a => Map Text (Expr Src a) -> Doc Ann
     prettyRecord =
         braces . map (prettyKeyValue colon) . Dhall.Map.toList
 
-    prettyRecordLit :: Pretty a => Map Text (Expr s a) -> Doc Ann
+    prettyRecordLit :: Pretty a => Map Text (Expr Src a) -> Doc Ann
     prettyRecordLit a
         | Data.Foldable.null a =
             lbrace <> equals <> rbrace
         | otherwise
             = braces (map (prettyKeyValue equals) (Dhall.Map.toList a))
 
-    prettyUnion :: Pretty a => Map Text (Expr s a) -> Doc Ann
+    prettyAlternative (key, Just val) = prettyKeyValue colon (key, val)
+    prettyAlternative (key, Nothing ) = duplicate (prettyAnyLabel key)
+
+    prettyUnion :: Pretty a => Map Text (Maybe (Expr Src a)) -> Doc Ann
     prettyUnion =
-        angles . map (prettyKeyValue colon) . Dhall.Map.toList
+        angles . map prettyAlternative . Dhall.Map.toList
 
-    prettyUnionLit
-        :: Pretty a => Text -> Expr s a -> Map Text (Expr s a) -> Doc Ann
-    prettyUnionLit a b c =
-        angles (front : map adapt (Dhall.Map.toList c))
-      where
-        front = prettyKeyValue equals (a, b)
-
-        adapt = prettyKeyValue colon
-
-    prettyChunks :: Pretty a => Chunks s a -> Doc Ann
+    prettyChunks :: Pretty a => Chunks Src a -> Doc Ann
     prettyChunks (Chunks a b) =
         if any (\(builder, _) -> hasNewLine builder) a || hasNewLine b
         then Pretty.flatAlt long short
